import numpy as np
from skimage.io import imread

import pyclesperanto as cle

cle.select_device("TX")


def test_connected_components_labeling_box():
    gpu_input = cle.push(np.asarray([[[1, 0, 1], [1, 0, 0], [0, 0, 1]]]))

    gpu_reference = cle.push(np.asarray([[[1, 0, 2], [1, 0, 0], [0, 0, 3]]]))

    gpu_output = cle.connected_components_labeling(gpu_input, connectivity="box")

    a = cle.pull(gpu_output)
    b = cle.pull(gpu_reference)

    print(a)
    print(b)

    assert np.array_equal(a, b)


def test_connected_components_labeling_sphere():
    gpu_input = cle.push(np.asarray([[0, 1, 0, 1], [0, 1, 0, 0], [1, 0, 0, 1]]))

    gpu_reference = cle.push(np.asarray([[0, 1, 0, 2], [0, 1, 0, 0], [3, 0, 0, 4]]))

    gpu_output = cle.connected_components_labeling(gpu_input, connectivity="sphere")

    a = cle.pull(gpu_output)
    b = cle.pull(gpu_reference)

    print(b)
    print(a)

    assert np.array_equal(a, b)


def test_connected_components_labeling_box_blobs():
    # initialize GPU
    print("Used GPU: " + cle.get_device().name)

    # load data
    filename = "https://samples.fiji.sc/blobs.png"
    image = np.squeeze(imread(filename))
    print("Loaded image size: " + str(image.shape))

    # push image to GPU memory
    input = cle.push(image)
    print("Image size in GPU: " + str(input.shape))

    # process the image
    blurred = cle.gaussian_blur(image, sigma_x=1, sigma_y=1)
    binary = cle.threshold_otsu(blurred)
    labeled = cle.connected_components_labeling(binary, connectivity="box")

    # The maxmium intensity in a label image corresponds to the number of objects
    num_labels = cle.maximum_of_all_pixels(labeled)

    # print out result
    print("Num objects in the image: " + str(num_labels))

    assert num_labels == 62



import numpy as np
from skimage.data import camera
from skimage.filters import threshold_otsu

import pyclesperanto as cle

cle.select_device("TX")


def test_threshold_otsu_against_scikit_image():
    image = camera()
    thresh = threshold_otsu(image)
    binary = image > thresh

    print(thresh)

    # from skimage import exposure
    # counts, bin_centers = exposure.histogram(image.ravel(), 256, source_range='image')

    # print(str(counts))
    # print(str(bin_centers))

    # threshold in GPU

    gpu_image = cle.push(image)
    gpu_binary = cle.threshold_otsu(gpu_image)

    print(str(binary))
    print(str(cle.pull(gpu_binary)))

    # compare
    assert np.allclose(binary, (cle.pull(gpu_binary) > 0))



import numpy as np

import pyclesperanto as cle


def test_mod():
    a = np.array([1, 2, 3, 4, 5]).astype(float)
    b = np.array([2, 2, 2, 2, 2]).astype(float)
    c = cle.mod(a, b)

    res = cle.pull(c)

    assert (res == np.array([1, 0, 1, 0, 1])).all()


def test_sqrt():
    a = np.array([1, 9, 25]).astype(float)
    c = cle.sqrt(a)

    res = cle.pull(c)

    assert (res == np.array([1, 3, 5])).all()


def test_cbrt():
    a = np.array([1, 27, 125]).astype(float)
    c = cle.cbrt(a)

    res = cle.pull(c)

    assert (res == np.array([1, 3, 5])).all()


def test_fabs():
    a = np.array([-1, -2, -3, -4, -5]).astype(float)
    c = cle.fabs(a)

    res = cle.pull(c)

    assert (res == np.array([1, 2, 3, 4, 5])).all()


def test_power_scalar():
    a = np.array([1, 2, 3, 4, 5]).astype(float)
    c = cle.power(a, 3)

    res = cle.pull(c)

    assert (res == np.array([1, 8, 27, 64, 125])).all()


def test_power_images():
    a = np.array([1, 2, 3, 4, 5]).astype(float)
    b = np.array([3, 3, 3, 3, 3]).astype(float)

    c = cle.power(a, b)
    res = cle.pull(c)

    assert (res == np.array([1, 8, 27, 64, 125])).all()



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_np_asarray():
    data = np.asarray([[[1.0, 2.0, 3.0], [1.0, 2.0, 3.0]]]).astype(np.float32)

    cl_data = cle.push(data)

    assert data.dtype == cl_data.dtype

    assert data.dtype == np.asarray(cl_data).dtype

    labels = cle.voronoi_otsu_labeling(cl_data)
    print(labels.dtype)
    print(np.asarray(labels).dtype)

    assert labels.dtype == np.asarray(labels).dtype



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_logarithm():
    test1 = cle.push(np.asarray([[1, 10], [1000, 100]]))

    reference = cle.push(np.asarray([[0, 2.3026], [6.9078, 4.6052]]))

    result = cle.create(test1, dtype=float)
    cle.logarithm(test1, result)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)
    print(b)

    assert np.allclose(a, b, 0.01)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_maximum_image_and_scalar():
    test1 = cle.push(np.asarray([[0, 3, 4, 5, 0], [0, 2, 1, 6, 0], [0, 0, 8, 7, 0]]))

    reference = cle.push(
        np.asarray([[2, 3, 4, 5, 2], [2, 2, 2, 6, 2], [2, 2, 8, 7, 2]])
    )

    result = cle.create(test1)
    cle.maximum_image_and_scalar(test1, result, 2)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)

    assert np.array_equal(a, b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_ellipsis():
    image = cle.asarray(
        [
            [[1, 2], [3, 4]],
            [[5, 6], [7, 8]],
        ]
    )
    print(image.shape, type(image))
    np_image = np.asarray(
        [
            [[1, 2], [3, 4]],
            [[5, 6], [7, 8]],
        ]
    )
    print(image.shape, type(image))

    assert np.array_equal(image[1, ...], np_image[1, ...])
    assert np.array_equal(image[:, 1, ...], np_image[:, 1, ...])
    assert np.array_equal(image[..., 0], np_image[..., 0])
    assert np.array_equal(image[0, ..., 0], np_image[0, ..., 0])
    assert np.array_equal(image[0, ..., :], np_image[0, ..., :])



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_t_2d():
    image = np.random.random((3, 2))

    cle_image = cle.asarray(image)

    assert cle.array_equal(cle_image.T, image.T)


def test_t_3d():
    image = np.random.random((3, 2, 5))

    cle_image = cle.asarray(image)

    assert cle.array_equal(cle_image.T, image.T)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_getitem_3d():
    data = np.asarray([[[1.0, 2.0, 3.0], [1.0, 2.0, 3.0]]]).astype(np.float32)
    ref = np.asarray([[1.0, 2.0, 1.0]]).astype(np.float32)

    cl_data = cle.push(data)

    positions = (np.asarray([0, 0, 0]), np.asarray([0, 1, 1]), np.asarray([0, 1, 0]))

    values = cl_data[positions]

    assert np.allclose(values, ref)


def test_getitem_2d():
    data = np.asarray([[1.0, 2.0, 3.0], [1.0, 2.0, 3.0]]).astype(np.float32)
    ref = np.asarray([[[1.0, 2.0, 2.0]]]).astype(np.float32)

    cl_data = cle.push(data)

    positions = (np.asarray([0, 1, 0]), np.asarray([0, 1, 1]))

    values = cl_data[positions]

    assert np.allclose(values, ref)


def test_clamp_to_edge():
    x = cle.push([[np.arange(10)]])
    print(x.shape)

    positions = (np.asarray([0]), np.asarray([0]), np.asarray([20]))

    assert x[positions] == 9


def test_tuple():
    x = cle.push([[np.arange(10)]])
    print(x.shape)

    positions = (np.asarray([0]), np.asarray([0]), np.asarray([4]))

    assert x[positions] == 4


def test_tuple2():
    x = cle.push([[np.arange(10)]])
    print(x.shape)

    positions = (0, 0, 4)

    assert x[positions] == 4


def test_list():
    x = cle.push([[np.arange(10)]])
    print(x.shape)

    positions = [0, 0, 4]

    assert x[positions] == 4


def test_list2():
    x = cle.push([[np.arange(10)]])
    print(x.shape)

    positions = [[0, 0], [0, 0], [4, 5]]

    assert np.allclose(x[positions], [4, 5])


def test_list_tuple_mix():
    x = cle.push([[np.arange(10)]])
    print(x.shape)

    positions = [(0, 0), (0, 0), (4, 5)]

    assert np.allclose(x[positions], [4, 5])


def test_list_tuple_mix2():
    x = cle.push([[np.arange(10)]])
    print(x.shape)

    positions = ([0, 0], [0, 0], [4, 5])

    assert np.allclose(x[positions], [4, 5])



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_reciprocal():
    test = [[0.2, 0.1], [10, 20]]
    reference = [[5, 10], [0.1, 0.05]]

    result = cle.reciprocal(test)

    assert np.allclose(result, reference)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_small_hessian_eigenvalue_2d():
    test = np.asarray([[1, -1], [1, -1]])

    reference_small_hessian_eigenvalue = np.asarray([[-2, 0], [-2, 0]])

    small_hessian_eigenvalue = cle.small_hessian_eigenvalue(test)

    print(small_hessian_eigenvalue)

    assert np.allclose(reference_small_hessian_eigenvalue, small_hessian_eigenvalue)


def test_small_hessian_eigenvalue_3d():
    test = np.asarray(
        [
            [[1, -1], [1, -1]],
            [[2, -2], [2, -2]],
        ]
    )

    reference_small_hessian_eigenvalue = np.asarray(
        [
            [[-2.1, -1.1], [-2.1, -1.1]],
            [[-4.1, 0], [-4.1, 0]],
        ]
    )

    small_hessian_eigenvalue = cle.small_hessian_eigenvalue(test)

    print(small_hessian_eigenvalue)

    assert np.allclose(
        reference_small_hessian_eigenvalue, small_hessian_eigenvalue, atol=0.1
    )



import numpy as np

cl_buffer_datatype_dict = {
    bool: "bool",
    np.uint8: "uchar",
    np.uint16: "ushort",
    np.uint32: "uint",
    np.uint64: "ulong",
    np.int8: "char",
    np.int16: "short",
    np.int32: "int",
    np.int64: "long",
    np.float32: "float",
    np.complex64: "cfloat_t",
    int: "int",
    float: "float",
    np.float64: "float",
}

_supported_numeric_types = tuple(cl_buffer_datatype_dict.keys())


def _correct_range(start, stop, step, size):
    # set in case not set (passed None)
    if step is None:
        step = 1
    if start is None:
        if step >= 0:
            start = 0
        else:
            start = size - 1

    if stop is None:
        if step >= 0:
            stop = size
        else:
            stop = -1

    # Check if ranges make sense
    if start >= size:
        if step >= 0:
            start = size
        else:
            start = size - 1
    if start < -size + 1:
        start = -size + 1
    if stop > size:
        stop = size
    if stop < -size:
        if start > 0:
            stop = 0 - 1
        else:
            stop = -size

    if start < 0:
        start = size - start
    if (start > stop and step > 0) or (start < stop and step < 0):
        # swap start and stop
        start, stop = stop, start

    return start, stop, step


def _compute_range(key, shape):
    use_range = any([(index[2] != 1 and index[2]) for index in key])

    range_x = _correct_range(key[-1][0], key[-1][1], key[-1][2], shape[-1])
    range_y = (
        _correct_range(key[-2][0], key[-2][1], key[-2][2], shape[-2])
        if len(shape) > 1
        else [0, 1, 1]
    )
    range_z = (
        _correct_range(key[-3][0], key[-3][1], key[-3][2], shape[-3])
        if len(shape) > 2
        else [0, 1, 1]
    )

    return use_range, range_x, range_y, range_z


def _process_ellipsis_into_slice(index, shape):
    if Ellipsis in index:
        ellipsis_index = index.index(Ellipsis)
        num_slices = len(shape) - (len(index) - 1)
        slices = (slice(None, None, None),) * num_slices
        index = index[:ellipsis_index] + slices + index[ellipsis_index + 1 :]
    return index


def _trim_index_to_shape(index, shape):
    if len(index) > len(shape):
        index = index[-len(shape) :]
    return index


def _assert_supported_dtype(x):
    x_type = getattr(x, "type", x)
    if x_type not in _supported_numeric_types:
        raise TypeError("dtype %s not supported " % x_type)



#
# This code is auto-generated from CLIc 'cle::tier3.hpp' file, do not edit manually.
#

import importlib
import warnings
from typing import Optional

import numpy as np

from ._array import Image
from ._core import Device
from ._decorators import plugin_function

clic = importlib.import_module('._pyclesperanto', package='pyclesperanto')

@plugin_function
def bounding_box(
    input_image: Image,
    device: Optional[Device] =None
) -> list:
    """Determines the bounding box of all nonzero pixels in a binary image. The
    positions are returned in  an array of 6 values as follows: minX, minY, minZ,
    maxX, maxY, maxZ.

    Parameters
    ----------
    input_image: Image 
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    list

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_boundingBox
    """
    return clic._bounding_box(device, input_image)

@plugin_function
def center_of_mass(
    input_image: Image,
    device: Optional[Device] =None
) -> list:
    """Determines the center of mass of an image or image stack. It writes the result
    in the results table in the columns MassX, MassY and MassZ.

    Parameters
    ----------
    input_image: Image 
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    list

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_centerOfMass
    """
    return clic._center_of_mass(device, input_image)

@plugin_function
def remove_labels(
    input_image: Image,
    list: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """This operation removes labels from a labelmap and renumbers the remaining
    labels. Hand over a binary flag list vector starting with a flag for the
    background, continuing with label1, label2,... For example if you pass
    0,1,0,0,1: Labels 1 and 4 will be removed (those with a 1 in the vector will be
    excluded). Labels 2 and 3 will be kept and renumbered to 1 and 2.

    Parameters
    ----------
    input_image: Image 
        
    list: Image 
        
    output_image: Optional[Image] (= None)
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_excludeLabels
    """
    return clic._remove_labels(device, input_image, list, output_image)

@plugin_function
def exclude_labels(
    input_image: Image,
    list: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """This operation removes labels from a labelmap and renumbers the remaining
    labels. Hand over a binary flag list vector starting with a flag for the
    background, continuing with label1, label2,... For example if you pass
    0,1,0,0,1: Labels 1 and 4 will be removed (those with a 1 in the vector will be
    excluded). Labels 2 and 3 will be kept and renumbered to 1 and 2.

    Parameters
    ----------
    input_image: Image 
        
    list: Image 
        
    output_image: Optional[Image] (= None)
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_excludeLabels
    """
    return clic._exclude_labels(device, input_image, list, output_image)

@plugin_function(categories=["label processing", "in assistant", "bia-bob-suggestion"])
def remove_labels_on_edges(
    input_image: Image,
    output_image: Optional[Image] =None,
    exclude_x: bool =True,
    exclude_y: bool =True,
    exclude_z: bool =True,
    device: Optional[Device] =None
) -> Image:
    """Removes all labels from a label map which touch the edges of the image.
    Remaining label elements are renumbered afterwards.

    Parameters
    ----------
    input_image: Image 
        
    output_image: Optional[Image] (= None)
        
    exclude_x: bool (= True)
        Exclude labels along min and max x
    exclude_y: bool (= True)
        Exclude labels along min and max y
    exclude_z: bool (= True)
        Exclude labels along min and max z
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_excludeLabelsOnEdges
    """
    return clic._remove_labels_on_edges(device, input_image, output_image, exclude_x, exclude_y, exclude_z)

@plugin_function(categories=["label processing", "in assistant"])
def exclude_labels_on_edges(
    input_image: Image,
    output_image: Optional[Image] =None,
    exclude_x: bool =True,
    exclude_y: bool =True,
    exclude_z: bool =True,
    device: Optional[Device] =None
) -> Image:
    """Removes all labels from a label map which touch the edges of the image.
    Remaining label elements are renumbered afterwards.

    Parameters
    ----------
    input_image: Image 
        
    output_image: Optional[Image] (= None)
        
    exclude_x: bool (= True)
        Exclude labels along min and max x
    exclude_y: bool (= True)
        Exclude labels along min and max y
    exclude_z: bool (= True)
        Exclude labels along min and max z
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_excludeLabelsOnEdges
    """
    return clic._exclude_labels_on_edges(device, input_image, output_image, exclude_x, exclude_y, exclude_z)

@plugin_function
def flag_existing_labels(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Given a label map this function will generate a binary vector where all pixels
    are set to 1 if label with given xcoordinate in the vector exists. For example a
    label image such as ``` 0 1 3 5 ``` will produce a flag_vector like this: ``` 1
    1 0 1 0 1 ```

    Parameters
    ----------
    input_image: Image 
        a label image
    output_image: Optional[Image] (= None)
        binary vector, if given should have size 1*n with n = maximum label + 1
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._flag_existing_labels(device, input_image, output_image)

@plugin_function(categories=["filter", "in assistant"])
def gamma_correction(
    input_image: Image,
    output_image: Optional[Image] =None,
    gamma: float =1,
    device: Optional[Device] =None
) -> Image:
    """Applies a gamma correction to an image. Therefore, all pixels x of the Image X
    are normalized and the power to gamma g is computed, before normlization is
    reversed (^ is the power operator):f(x) = (x / max(X)) ^ gamma * max(X)

    Parameters
    ----------
    input_image: Image 
        
    output_image: Optional[Image] (= None)
        
    gamma: float (= 1)
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_gammaCorrection
    """
    return clic._gamma_correction(device, input_image, output_image, float(gamma))

@plugin_function
def generate_binary_overlap_matrix(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Takes two labelmaps with n and m labels and generates a (n+1)*(m+1) matrix where
    all pixels are set to 0 exept those where labels overlap between the label maps.
    For example, if labels 3 in labelmap1 and 4 in labelmap2 are touching then the
    pixel (3,4) in the matrix will be set to 1.

    Parameters
    ----------
    input_image0: Image 
        
    input_image1: Image 
        
    output_image: Optional[Image] (= None)
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_generateBinaryOverlapMatrix
    """
    return clic._generate_binary_overlap_matrix(device, input_image0, input_image1, output_image)

@plugin_function
def generate_touch_matrix(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Takes a labelmap with n labels and generates a (n+1)*(n+1) matrix where all
    pixels are set to 0 exept those where labels are touching. Only half of the
    matrix is filled (with x < y). For example, if labels 3 and 4 are touching then
    the pixel (3,4) in the matrix will be set to 1. The touch matrix is a
    representation of a region adjacency graph

    Parameters
    ----------
    input_image: Image 
        
    output_image: Optional[Image] (= None)
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_generateTouchMatrix
    """
    return clic._generate_touch_matrix(device, input_image, output_image)

@plugin_function
def histogram(
    input_image: Image,
    output_image: Optional[Image] =None,
    nbins: int =256,
    min: Optional[float] =None,
    max: Optional[float] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines the histogram of a given image. The histogram image is of dimensions
    number_of_bins/1/1; a 3D image with height=1 and depth=1. Histogram bins contain
    the number of pixels with intensity in this corresponding bin. The histogram
    bins are uniformly distributed between given minimum and maximum grey value
    intensity. If the flag determine_min_max is set, minimum and maximum intensity
    will be determined. When calling this operation many times, it is recommended to
    determine minimum and maximum intensity once at the beginning and handing over
    these values. Author(s): Robert Haase adapted work from Aaftab Munshi, Benedict
    Gaster, Timothy Mattson, James Fung, Dan Ginsburg License: adapted code from
    https://github.com/bgaster/openclbooksamples/blob/master/src/Chapter_14/histogram/histogram_image.cl
    It was published unter BSD license according to
    https://code.google.com/archive/p/openclbooksamples/ Book: OpenCL(R) Programming
    Guide Authors: Aaftab Munshi, Benedict Gaster, Timothy Mattson, James Fung, Dan
    Ginsburg ISBN10: 0321749642 ISBN13: 9780321749642 Publisher: AddisonWesley
    Professional URLs: http://safari.informit.com/9780132488006/
    http://www.openclprogrammingguide.com

    Parameters
    ----------
    input_image: Image 
        
    output_image: Optional[Image] (= None)
        
    nbins: int (= 256)
        
    min: Optional[float] (= None)
        
    max: Optional[float] (= None)
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_histogram
    """
    return clic._histogram(device, input_image, output_image, int(nbins), min, max)

@plugin_function
def jaccard_index(
    input_image0: Image,
    input_image1: Image,
    device: Optional[Device] =None
) -> float:
    """Determines the overlap of two binary images using the Jaccard index. A value of
    0 suggests no overlap, 1 means perfect overlap. The resulting Jaccard index is
    saved to the results table in the 'Jaccard_Index' column. Note that the
    SorensenDice coefficient can be calculated from the Jaccard index j using this
    formula: <pre>s = f(j) = 2 j / (j + 1)</pre>

    Parameters
    ----------
    input_image0: Image 
        
    input_image1: Image 
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    float

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_jaccardIndex
    """
    return clic._jaccard_index(device, input_image0, input_image1)

@plugin_function
def labelled_spots_to_pointlist(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Generates a coordinate list of points in a labelled spot image. Transforms a
    labelmap of spots (single pixels with values 1, 2,..., n for n spots) as
    resulting from connected components analysis in an image where every column
    contains d pixels (with d = dimensionality of the original image) with the
    coordinates of the maxima/minima.

    Parameters
    ----------
    input_image: Image 
        
    output_image: Optional[Image] (= None)
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_labelledSpotsToPointList
    """
    return clic._labelled_spots_to_pointlist(device, input_image, output_image)

@plugin_function
def maximum_position(
    input_image: Image,
    device: Optional[Device] =None
) -> list:
    """Determines the position of the maximum of all pixels in a given image.

    Parameters
    ----------
    input_image: Image 
        The image of which the position of the maximum of all pixels will be determined.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    list
    """
    return clic._maximum_position(device, input_image)

@plugin_function
def mean_of_all_pixels(
    input_image: Image,
    device: Optional[Device] =None
) -> float:
    """Determines the mean average of all pixels in a given image.

    Parameters
    ----------
    input_image: Image 
        The image of which the mean average of all pixels will be determined.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    float

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_meanOfAllPixels
    """
    return clic._mean_of_all_pixels(device, input_image)

@plugin_function
def minimum_position(
    input_image: Image,
    device: Optional[Device] =None
) -> list:
    """Determines the position of the minimum of all pixels in a given image.

    Parameters
    ----------
    input_image: Image 
        The image of which the position of the minimum of all pixels will be determined.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    list
    """
    return clic._minimum_position(device, input_image)

@plugin_function
def morphological_chan_vese(
    input_image: Image,
    output_image: Optional[Image] =None,
    num_iter: int =100,
    smoothing: int =1,
    lambda1: float =1,
    lambda2: float =1,
    device: Optional[Device] =None
) -> Image:
    """Compute an active contour model using the Chan-Vese morphological algorithm. The
    output image (dst) should also be initialisation of the contour. If not provided
    (nullptr), the function will use a checkboard pattern initialisation.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output contour, can also be use to provide initialisation.
    num_iter: int (= 100)
        Number of iterations.
    smoothing: int (= 1)
        Number of
    lambda1: float (= 1)
        Lambda1.
    lambda2: float (= 1)
        Lambda2.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._morphological_chan_vese(device, input_image, output_image, int(num_iter), int(smoothing), float(lambda1), float(lambda2))

@plugin_function
def statistics_of_labelled_pixels(
    label: Image,
    intensity: Optional[Image] =None,
    device: Optional[Device] =None
) -> dict:
    """Compute the bounding box, area (in pixels/voxels), minimum intensity, maximum
    intensity, average intensity, standard deviation of the intensity, and some
    shape descriptors of labelled objects in a label image and its corresponding
    intensity image. The intensity image is optional and set to 0 if not provided.

    Parameters
    ----------
    label: Image 
        Label image to compute the statistics.
    intensity: Optional[Image] (= None)
        Intensity image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    dict

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_statisticsOfLabelledPixels
    """
    return clic._statistics_of_labelled_pixels(device, label, intensity)

@plugin_function
def statistics_of_background_and_labelled_pixels(
    label: Image,
    intensity: Optional[Image] =None,
    device: Optional[Device] =None
) -> dict:
    """Compute, for the background and labels, the bounding box, area (in
    pixels/voxels), minimum intensity, maximum intensity, average intensity,
    standard deviation of the intensity, and some shape descriptors of labelled
    objects in a label image and its corresponding intensity image. The intensity
    image is optional and set to 0 if not provided.

    Parameters
    ----------
    label: Image 
        Label image to compute the statistics.
    intensity: Optional[Image] (= None)
        Intensity image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    dict

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_statisticsOfBackgroundAndLabelledPixels
    """
    return clic._statistics_of_background_and_labelled_pixels(device, label, intensity)


import numpy as np
from skimage import morphology, segmentation

import pyclesperanto as cle

# def test_morphological_chan_vese():
#     image = generate_disk((25, 25), 5).astype(np.float32)
#     reference = cle.push(segmentation.morphological_chan_vese(image, num_iter=1, smoothing=1, lambda1=1, lambda2=1))
#     result = cle.morphological_chan_vese(image, num_iter=1, smoothing=, lambda1=1, lambda2=1)
#     a = cle.pull(result).astype(np.uint8)
#     b = cle.pull(reference).astype(np.uint8)
#     print(a)
#     print(b)
#     assert (np.array_equal(a, b))


# def test_morphological_chan_vese_on_membranes_2d():
#     from skimage.data import cells3d
#     from skimage.segmentation import morphological_chan_vese
#     import pyclesperanto as cle
#     image2d = cells3d()[30, 0, ...]
#     result_gpu = cle.morphological_chan_vese(image2d, num_iter=20, smoothing=10)
#     result_cpu = morphological_chan_vese(image2d, num_iter=20, smoothing=10)
#     assert cle.array_equal(result_cpu, result_gpu)

# def test_morphological_chan_vese_on_membranes_3d():
#     from skimage.data import cells3d
#     from skimage.segmentation import morphological_chan_vese
#     import pyclesperanto as cle
#     image3d = cells3d()[10:50:, 0, :40, :40]
#     result_gpu = cle.morphological_chan_vese(image3d, num_iter=20, smoothing=10)
#     result_cpu = morphological_chan_vese(image3d, num_iter=20, smoothing=10)
#     cle.array_equal(result_cpu, result_gpu)


def generate_disk(shape, radius):
    image = np.zeros(shape)
    image[
        image.shape[0] // 2 - radius : image.shape[0] // 2 + radius + 1,
        image.shape[1] // 2 - radius : image.shape[1] // 2 + radius + 1,
    ] = morphology.disk(radius)
    return image



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_top_hat_sphere():
    test = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 50, 50, 50, 0],
                [0, 50, 100, 50, 0],
                [0, 50, 50, 50, 0],
                [0, 0, 0, 0, 0],
            ]
        )
    )

    result = cle.create(test)
    cle.top_hat(test, result, 1, 1, 0, connectivity="sphere")

    print(result)

    a = cle.pull(result)
    assert np.min(a) == 0
    assert np.max(a) == 50



#
# This code is auto-generated from CLIc 'cle::tier5.hpp' file, do not edit manually.
#

import importlib
import warnings
from typing import Optional

import numpy as np

from ._array import Image
from ._core import Device
from ._decorators import plugin_function

clic = importlib.import_module('._pyclesperanto', package='pyclesperanto')

@plugin_function(categories=["combine"])
def array_equal(
    input_image0: Image,
    input_image1: Image,
    device: Optional[Device] =None
) -> bool:
    """Compares if all pixels of two images are identical. If shape of the images or
    any pixel are different, returns False. True otherwise This function is supposed
    to work similarly like its counterpart in numpy [1].

    Parameters
    ----------
    input_image0: Image 
        
    input_image1: Image 
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    bool

    References
    ----------
    [1] https://numpy.org/doc/stable/reference/generated/numpy.array_equal.html
    """
    return clic._array_equal(device, input_image0, input_image1)

@plugin_function(categories=["label processing", "combine labels", "in assistant", "bia-bob-suggestion"])
def combine_labels(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Combines two label images by adding labels of a given label image to another.
    Labels in the second image overwrite labels in the first passed image.
    Afterwards, labels are relabeled sequentially.

    Parameters
    ----------
    input_image0: Image 
        label image to add labels to.
    input_image1: Image 
        label image to add labels from.
    output_image: Optional[Image] (= None)
        Output label image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._combine_labels(device, input_image0, input_image1, output_image)

@plugin_function(categories=["label", "in assistant"])
def connected_components_labeling(
    input_image: Image,
    output_image: Optional[Image] =None,
    connectivity: str ='box',
    device: Optional[Device] =None
) -> Image:
    """Performs connected components analysis inspecting the box neighborhood of every
    pixel to a binary image and generates a label map.

    Parameters
    ----------
    input_image: Image 
        Binary image to label.
    output_image: Optional[Image] (= None)
        Output label image.
    connectivity: str (= 'box')
        Defines pixel neighborhood relationship, "box" or "sphere".
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_connectedComponentsLabelingBox
    """
    return clic._connected_components_labeling(device, input_image, output_image, str(connectivity))

@plugin_function(categories=["label", "in assistant", "bia-bob-suggestion"])
def connected_component_labeling(
    input_image: Image,
    output_image: Optional[Image] =None,
    connectivity: str ='box',
    device: Optional[Device] =None
) -> Image:
    """Performs connected components analysis inspecting the box neighborhood of every
    pixel to a binary image and generates a label map.

    Parameters
    ----------
    input_image: Image 
        Binary image to label.
    output_image: Optional[Image] (= None)
        Output label image.
    connectivity: str (= 'box')
        Defines pixel neighborhood relationship, "box" or "sphere".
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_connectedComponentsLabelingBox
    """
    return clic._connected_component_labeling(device, input_image, output_image, str(connectivity))

@plugin_function(categories=["label processing", "in assistant", "bia-bob-suggestion"])
def reduce_labels_to_centroids(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Take a label map and reduce each label to its centroid.

    Parameters
    ----------
    input_image: Image 
        Label image to reduce.
    output_image: Optional[Image] (= None)
        Output label image with centroids.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_reduceLabelsToCentroids
    """
    return clic._reduce_labels_to_centroids(device, input_image, output_image)

@plugin_function(categories=["label processing", "in assistant"])
def filter_label_by_size(
    input_image: Image,
    output_image: Optional[Image] =None,
    min_size: float =0,
    max_size: float =100,
    device: Optional[Device] =None
) -> Image:
    """Filter labelled objects outside of the min/max size range value.

    Parameters
    ----------
    input_image: Image 
        Input label image.
    output_image: Optional[Image] (= None)
        Output label image.
    min_size: float (= 0)
        Minimum size of labels to keep.
    max_size: float (= 100)
        Maximum size of labels to keep.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_excludeLabelsOutsideSizeRange
    """
    return clic._filter_label_by_size(device, input_image, output_image, float(min_size), float(max_size))

@plugin_function(categories=["label processing", "in assistant"])
def exclude_labels_outside_size_range(
    input_image: Image,
    output_image: Optional[Image] =None,
    min_size: float =0,
    max_size: float =100,
    device: Optional[Device] =None
) -> Image:
    """Filter labelled objects outside of the min/max size range value.

    Parameters
    ----------
    input_image: Image 
        Input label image.
    output_image: Optional[Image] (= None)
        Output label image.
    min_size: float (= 0)
        Minimum size of labels to keep.
    max_size: float (= 100)
        Maximum size of labels to keep.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_excludeLabelsOutsideSizeRange
    """
    return clic._exclude_labels_outside_size_range(device, input_image, output_image, float(min_size), float(max_size))


import numpy as np
import pytest

import pyclesperanto as cle

cle.select_device("TX")

dtypes = {
    "int8",
    "int16",
    "int32",
    # 'int64',
    "uint8",
    "uint16",
    "uint32",
    # 'uint64',
    # 'float16',
    "float32",
    # 'float64',
    # 'complex64',
}


@pytest.fixture(params=dtypes)
def dtype(request):
    return np.dtype(request.param)


@pytest.mark.parametrize(
    "shape", [(256, 256), (3, 256, 256), (256, 256, 3), (10, 256, 256)]
)
def test_create_image(dtype, shape):
    array = np.random.randint(0, 255, shape).astype(dtype)
    cle.create(array, mtype="image")



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_set_image_borders():
    result = cle.push(
        np.asarray(
            [
                [3, 3, 3, 3, 3],
                [3, 3, 3, 3, 3],
                [3, 3, 3, 3, 3],
                [3, 3, 3, 3, 3],
                [3, 3, 3, 3, 3],
            ]
        )
    )

    reference = cle.push(
        np.asarray(
            [
                [4, 4, 4, 4, 4],
                [4, 3, 3, 3, 4],
                [4, 3, 3, 3, 4],
                [4, 3, 3, 3, 4],
                [4, 4, 4, 4, 4],
            ]
        )
    )

    cle.set_image_borders(result, 4)

    print(result)

    a = cle.pull(result)
    b = cle.pull(reference)
    assert np.allclose(a, b, 0.001)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_generate_distance_matrix():
    gpu_input = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 1, 0, 3, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 2, 0, 0],
                [0, 0, 0, 0, 4],
            ]
        )
    )

    gpu_reference = cle.push(
        np.asarray(
            [
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 2.236068, 2.0, 4.2426405],
                [0.0, 2.236068, 0.0, 2.236068, 2.236068],
                [0.0, 2.0, 2.236068, 0.0, 3.1622777],
                [0.0, 4.2426405, 2.236068, 3.1622777, 0.0],
            ]
        )
    )

    gpu_pointlist = cle.labelled_spots_to_pointlist(gpu_input)
    gpu_distance_matrix = cle.generate_distance_matrix(gpu_pointlist, gpu_pointlist)

    a = cle.pull(gpu_distance_matrix)
    b = cle.pull(gpu_reference)

    print(a)
    print(b)

    assert np.allclose(a, b, 0.001)



import numpy as np

import pyclesperanto as cle


def test_concate_along_x():
    test1 = cle.push(np.asarray([[1, 1], [1, 1]]))
    test2 = cle.push(np.asarray([[2, 2, 2], [2, 2, 2]]))

    reference = cle.push(np.asarray([[1, 1, 2, 2, 2], [1, 1, 2, 2, 2]]))

    result = cle.concatenate_along_x(test1, test2)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)
    print(b)

    assert np.allclose(a, b, 0.01)


def test_concate_along_y():
    test1 = cle.push(np.asarray([[1, 1], [1, 1]]))
    test2 = cle.push(np.asarray([[2, 2], [2, 2]]))

    reference = cle.push(np.asarray([[1, 1], [1, 1], [2, 2], [2, 2]]))

    result = cle.concatenate_along_y(test1, test2)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)
    print(b)

    assert np.allclose(a, b, 0.01)


def test_concate_along_z():
    test1 = cle.push(
        np.asarray(
            [
                [
                    [1, 1],
                ],
                [[1, 1]],
            ]
        )
    )
    test2 = cle.push(np.asarray([[[2, 2]], [[2, 2]], [[2, 2]]]))

    reference = cle.push(
        np.asarray(
            [
                [
                    [1, 1],
                ],
                [
                    [1, 1],
                ],
                [
                    [2, 2],
                ],
                [[2, 2]],
                [[2, 2]],
            ]
        )
    )

    result = cle.concatenate_along_z(test1, test2)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)
    print(b)

    assert np.allclose(a, b, 0.01)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_create_3d():
    size = [2, 3, 4]

    image = cle.create(size)

    assert image.shape[0] == 2
    assert image.shape[1] == 3
    assert image.shape[2] == 4

    image2 = cle.create_like(image)
    assert image2.shape[0] == 2
    assert image2.shape[1] == 3
    assert image2.shape[2] == 4


def test_create_2d():
    size = [2, 3]

    image = cle.create(size)

    assert image.shape[0] == 2
    assert image.shape[1] == 3

    image2 = cle.create_like(image)
    assert image2.shape[0] == 2
    assert image2.shape[1] == 3


def test_create_uint8():
    image = cle.push([[-1, 1.5], [2000, -7.8]])
    reference = np.asarray([[0, 1], [255, 0]])

    target = cle.create(image.shape, dtype=np.uint8)

    cle.copy(image, target)

    print(target)

    assert np.allclose(target, reference)


def test_create_uint16():
    image = cle.push([[-1, 1.5], [2000, -7.8]])
    reference = np.asarray([[0, 1], [2000, 0]])

    target = cle.create(image.shape, dtype=np.uint16)

    cle.copy(image, target)

    print(target)

    assert np.allclose(target, reference)


def test_create_uint32():
    image = cle.push([[-1, 1.5], [2000, -7.8]])
    reference = np.asarray([[0, 1], [2000, 0]])

    target = cle.create(image.shape, dtype=np.uint32)

    cle.copy(image, target)

    print(target)

    assert np.allclose(target, reference)


def test_create_uint64():
    image = cle.push([[-1, 1.5], [2000, -7.8]])
    reference = np.asarray([[0, 1], [2000, 0]])

    target = cle.create(image.shape, dtype=np.uint64)

    cle.copy(image, target)

    print(target)

    assert np.allclose(target, reference)


def test_create_int8():
    image = cle.push([[-1, 1.5], [2000, -7.8]])
    reference = np.asarray([[-1, 1], [127, -7]])

    target = cle.create(image.shape, dtype=np.int8)

    cle.copy(image, target)

    print(target)

    assert np.allclose(target, reference)


def test_create_int16():
    image = cle.push([[-1, 1.5], [2000, -7.8]])
    reference = np.asarray([[-1, 1], [2000, -7]])

    target = cle.create(image.shape, dtype=np.int16)

    cle.copy(image, target)

    print(target)

    assert np.allclose(target, reference)


def test_create_int32():
    image = cle.push([[-1, 1.5], [200000, -7.8]])
    reference = np.asarray([[-1, 1], [200000, -7]])

    target = cle.create(image.shape, dtype=np.int32)

    cle.copy(image, target)

    print(target)

    assert np.allclose(target, reference)


def test_create_int64():
    image = cle.push([[-1, 1.5], [2000, -7.8]])
    reference = np.asarray([[-1, 1], [2000, -7]])

    target = cle.create(image.shape, dtype=np.int64)

    cle.copy(image, target)

    print(target)

    assert np.allclose(target, reference)


def test_create_like_numpy():
    image = np.random.random((10, 20))

    cle_image = cle.create_like(image)
    cle.copy(image, cle_image)

    assert cle.array_equal(image, cle_image)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_add_images_weighted_missing_parameters():
    input1 = np.asarray([1, 2, 3])
    input2 = np.asarray([4, 5, 6])

    reference = np.asarray([5, 7, 9])
    output = cle.add_images_weighted(input1, input2, None, 1, 1)
    result = cle.pull(output)

    print(result)
    print(reference)
    assert np.array_equal(result, reference)


def test_add_images_weighted_none_output():
    input1 = np.asarray([1, 2, 3])
    input2 = np.asarray([4, 5, 6])

    reference = np.asarray([9, 12, 15])
    output = cle.add_images_weighted(input1, input2, None, 1, 2)
    result = cle.pull(output)

    print(result)
    print(reference)
    assert np.array_equal(result, reference)


def test_add_images_weighted_named_parameters():
    input1 = np.asarray([1, 2, 3])
    input2 = np.asarray([4, 5, 6])

    reference = np.asarray([9, 12, 15])
    output = cle.add_images_weighted(input1, input2, None, factor0=1, factor1=2)
    result = cle.pull(output)

    print(result)
    print(reference)
    assert np.array_equal(result, reference)


def test_add_images_weighted_wrong_parameter_order():
    input1 = np.asarray([1, 2, 3])
    input2 = np.asarray([4, 5, 6])

    reference = np.asarray([9, 12, 15])
    output = cle.add_images_weighted(input1, input2, factor0=1, factor1=2)
    result = cle.pull(output)

    print(result)
    print(reference)
    assert np.array_equal(result, reference)


def test_add_images_weighted_parameters_wrong_order_and_missing():
    input1 = np.asarray([1, 2, 3])
    input2 = np.asarray([4, 5, 6])

    reference = np.asarray([9, 12, 15])
    output = cle.add_images_weighted(input1, input2, factor1=2, factor0=1)
    result = cle.pull(output)

    print(result)
    print(reference)
    assert np.array_equal(result, reference)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_extend_labeling_via_voronoi():
    gpu_input = cle.push(
        np.asarray(
            [
                [
                    [0, 0, 0, 0, 3, 3],
                    [0, 4, 0, 0, 3, 3],
                    [0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0],
                    [0, 2, 0, 0, 1, 0],
                    [0, 0, 0, 0, 0, 0],
                ]
            ]
        )
    )

    gpu_reference = cle.push(
        np.asarray(
            [
                [
                    [4, 4, 4, 3, 3, 3],
                    [4, 4, 4, 3, 3, 3],
                    [4, 4, 4, 3, 3, 3],
                    [2, 2, 2, 1, 1, 1],
                    [2, 2, 2, 1, 1, 1],
                    [2, 2, 2, 1, 1, 1],
                ]
            ]
        )
    )

    gpu_output = cle.extend_labeling_via_voronoi(gpu_input)

    a = cle.pull(gpu_output)
    b = cle.pull(gpu_reference)

    print(a)
    print(b)

    assert np.array_equal(a, b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_count_touching_neighbors():
    labels = cle.push(
        np.asarray(
            [
                [1, 1, 0, 3, 3],
                [1, 1, 2, 3, 3],
                [0, 2, 2, 2, 0],
                [4, 4, 2, 5, 5],
                [4, 4, 0, 5, 5],
            ]
        )
    )

    reference = cle.push(np.asarray([0, 1, 4, 1, 1, 1]))

    touch_matrix = cle.generate_touch_matrix(labels)

    neighbor_count_vector = cle.count_touching_neighbors(touch_matrix)

    a = cle.pull(neighbor_count_vector)
    b = cle.pull(reference)

    print(a)
    print(b)

    assert np.array_equal(a, b)


def test_count_touching_neighbors_not_ignoring_background():
    labels = cle.push(
        np.asarray(
            [
                [1, 1, 0, 3, 3],
                [1, 1, 2, 3, 3],
                [0, 2, 2, 2, 0],
                [4, 4, 2, 5, 5],
                [4, 4, 0, 5, 5],
            ]
        )
    )

    reference = cle.push(np.asarray([5, 2, 5, 2, 2, 2]))

    touch_matrix = cle.generate_touch_matrix(labels)

    neighbor_count_vector = cle.count_touching_neighbors(
        touch_matrix, ignore_background=False
    )

    a = cle.pull(neighbor_count_vector)
    b = cle.pull(reference)

    print(a)
    print(b)

    assert np.array_equal(a, b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_convolve():
    test = cle.push(np.asarray([[0, 0, 0], [0, 1, 0], [0, 0, 0]]))

    test1 = cle.push(np.asarray([[0, 1, 0], [1, 2, 1], [0, 1, 0]]))

    test2 = cle.create_like(test)
    cle.convolve(test, test1, test2)

    print(test2)

    a = cle.pull(test1)
    b = cle.pull(test2)

    assert np.allclose(np.min(a), np.min(b), atol=1e-6)
    assert np.allclose(np.max(a), np.max(b), atol=1e-6)
    assert np.allclose(np.mean(a), np.mean(b), atol=1e-6)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_grreater_or_equal():
    test = cle.push(
        np.asarray(
            [
                [0, -6, 0, -3, 0],
                [0, 1, 2, 3, 0],
            ]
        )
    )

    reference = cle.push(
        np.asarray(
            [
                [0, 6, 0, 3, 0],
                [0, -1, -2, -3, 0],
            ]
        )
    )

    result = cle.create(test)
    cle.invert(test, result)

    a = cle.pull(result)
    b = cle.pull(reference)
    print(a)

    assert np.array_equal(a, b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_cubic_root():
    test = np.asarray([[125, 64]])
    reference = np.asarray([[5, 4]])

    result = cle.cubic_root(test)

    assert np.allclose(result, reference)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_maximum_sphere_1():
    test = cle.push(np.asarray([[1, 1, 1], [1, 2, 1], [1, 1, 1]]))

    test2 = cle.create(test)
    cle.maximum(test, test2, 1, 1, 1, "sphere")

    a = cle.pull(test2)
    assert np.min(a) == 1
    assert np.max(a) == 2

    cle.set(test, 5)
    # print(test)
    # print(cle.pull(test))

    a = cle.pull(test)
    assert np.min(a) == 5
    assert np.max(a) == 5
    assert np.mean(a) == 5


def test_maximum_sphere_2():
    gpu_a = cle.push(np.asarray([[1, 1, 1], [1, 2, 1], [1, 1, 1]]))
    gpu_b = cle.create(gpu_a)
    cle.maximum(gpu_a, gpu_b, 1, 1, 1, "sphere")

    a = cle.pull(gpu_b)
    assert np.min(a) == 1
    assert np.max(a) == 2

    cle.set(gpu_a, 5)
    assert np.all(cle.pull(gpu_a) == 5)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_sum_of_all_pixels_3d():
    test1 = cle.push(
        np.asarray(
            [
                [
                    [0, 4, 0, 0, 2],
                    [0, 0, 0, 8, 0],
                    [3, 0, 0, 0, 0],
                    [0, 0, 0, 0, 1],
                    [0, 0, 2, 0, 0],
                ]
            ]
        )
    )

    s = cle.sum_of_all_pixels(test1)

    assert s == 20


def test_sum_of_all_pixels_2d():
    test1 = cle.push(
        np.asarray(
            [
                [0, 4, 0, 0, 2],
                [0, 0, 0, 8, 0],
                [3, 0, 0, 0, 0],
                [0, 0, 0, 0, 1],
                [0, 0, 2, 0, 0],
            ]
        )
    )

    s = cle.sum_of_all_pixels(test1)

    assert s == 20


def test_sum_of_all_pixels_1d():
    test1 = cle.push(np.asarray([0, 4, 0, 0, 2]))

    s = cle.sum_of_all_pixels(test1)

    assert s == 6


def test_sum_of_all_pixels_1d_y():
    test1 = cle.push(np.asarray([[0], [4], [0], [0], [2]]))

    s = cle.sum_of_all_pixels(test1)

    assert s == 6


def test_sum_of_all_pixels_1d_z():
    test1 = cle.push(np.asarray([[[0]], [[4]], [[0]], [[0]], [[2]]]))

    s = cle.sum_of_all_pixels(test1)

    assert s == 6



from ._array import Array, Image, is_image
from ._categories import categories
from ._core import (
    Device,
    default_initialisation,
    get_device,
    info,
    list_available_backends,
    list_available_devices,
    select_backend,
    select_device,
    wait_for_kernel_to_finish,
)
from ._functionalities import execute, imshow, native_execute, operation, operations
from ._memory import create, create_like, pull, push
from ._tier1 import *
from ._tier2 import *
from ._tier3 import *
from ._tier4 import *
from ._tier5 import *
from ._tier6 import *
from ._tier7 import *
from ._tier8 import *
from ._version import CLIC_VERSION as __clic_version__
from ._version import COMMON_ALIAS as __common_alias__
from ._version import VERSION as __version__

from ._interroperability import *  # isort:skip

default_initialisation()



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_set_where_x_equals_y():
    result = cle.push(
        np.asarray([[0, 0, 0, 1], [0, 0, 3, 1], [0, 0, 3, 1], [1, 1, 1, 1]])
    )

    reference = cle.push(
        np.asarray([[3, 0, 0, 1], [0, 3, 3, 1], [0, 0, 3, 1], [1, 1, 1, 3]])
    )

    cle.set_where_x_equals_y(result, 3)

    a = cle.pull(result)
    b = cle.pull(reference)
    print(a)

    assert np.array_equal(a, b)




def list_bia_bob_plugins():
    """List of function hints for bia_bob"""
    return """
    ## pyclesperanto
    pyclesperanto is a Python library for GPU-accelerated image processing and analysis. 
    To use it, you need to import it:
    ```
    import pyclesperanto as cle
    ```
    
    * Compute the absolute value of each pixel in an input image, optionally specifying an output image and device for processing.
    cle.absolute(input_image: ndarray) -> ndarray
    
    * Calculate the pixel-by-pixel absolute difference between two images, optionally storing the result in an output image and selecting a device for operation.
    cle.absolute_difference(input_image0: ndarray, input_image1: ndarray) -> ndarray
    
    * Generate a binary image by applying the binary AND operator to corresponding pixels of two input images, considering all non-zero pixels as 1.
    cle.binary_and(input_image0: ndarray, input_image1: ndarray) -> ndarray
    
    * Identify and highlight edge pixels of binary objects, setting them to 1 in the output image on a specified device.
    cle.binary_edge_detection(input_image: ndarray) -> ndarray
    
    * Convert an image to a binary image by negating all non-zero pixels using the binary NOT operator.
    cle.binary_not(input_image: ndarray) -> ndarray
    
    * Combine two images pixel-wise using the binary OR operator, interpreting any non-zero pixel values as 1, with optional specification of output image and device.
    cle.binary_or(input_image0: ndarray, input_image1: ndarray) -> ndarray
    
    * Subtract one binary image from another, optionally specifying an output image and device.
    cle.binary_subtract(input_image0: ndarray, input_image1: ndarray) -> ndarray
    
    * Generate a binary image by applying the XOR operation on two input images, interpreting all non-zero pixels as 1.
    cle.binary_xor(input_image0: ndarray, input_image1: ndarray) -> ndarray
    
    * Perform background subtraction using a bottomhat filter on an image with customizable region radii, connectivity, and processing device options.
    cle.bottom_hat(input_image: ndarray, radius_x: float = 1, radius_y: float = 1, radius_z: float = 1, connectivity: str = 'box') -> ndarray
    
    * Perform background subtraction on an input image using a bottomhat filter with customizable radius parameters for X, Y, and Z dimensions.
    cle.bottom_hat_sphere(input_image: ndarray, radius_x: float = 1, radius_y: float = 1, radius_z: float = 1) -> ndarray
    
    * calculate the centroids of all labels in a 3D label image and output the coordinates as a point list image, with optional control over background inclusion and device execution.
    cle.centroids_of_labels(input_image: ndarray, withBG: bool = False) -> ndarray
    
    * Apply spherical morphological closing to intensity or binary images with customizable radii and connectivity options, optionally specifying an output image and device.
    cle.closing(input_image: ndarray, radius_x: int = 1, radius_y: int = 1, radius_z: int = 0, connectivity: str = 'box') -> ndarray
    
    * Perform a morphological closing operation on a label image using an octagonal structuring element with optional radius and device specifications.
    cle.closing_labels(input_image: ndarray, radius: int = 0) -> ndarray
    
    * Apply morphological closing to images using a spherical footprint, adjustable by x, y, z radii and optional output and device specifications.
    cle.closing_sphere(input_image: ndarray, radius_x: int = 1, radius_y: int = 1, radius_z: int = 0) -> ndarray
    
    * Combine two label images by overwriting and sequentially relabeling, allowing specification of input images, output image, and device.
    cle.combine_labels(input_image0: ndarray, input_image1: ndarray) -> ndarray
    
    * Concatenate two images or image stacks along the X axis, with optional output image and device specifications.
    cle.concatenate_along_x(input_image0: ndarray, input_image1: ndarray) -> ndarray
    
    * Concatenate two images or stacks along the Y axis, specifying optional output and device parameters.
    cle.concatenate_along_y(input_image0: ndarray, input_image1: ndarray) -> ndarray
    
    * Concatenate two images or stacks along the Z axis to produce a combined output image, optionally specifying a device for the operation.
    cle.concatenate_along_z(input_image0: ndarray, input_image1: ndarray) -> ndarray
    
    * Generate a label map by performing connected components analysis on a binary image, considering either 'box' or 'sphere' pixel neighborhoods, with optional output image and device specification.
    cle.connected_component_labeling(input_image: ndarray, connectivity: str = 'box') -> ndarray
    
    * Analyze and label connected components in a binary image using specified neighborhood connectivity, with support for device-specific execution.
    cle.connected_components_labeling(input_image: ndarray, connectivity: str = 'box') -> ndarray
    
    * compute a vector indicating the number of touching neighbors for each label in a touch matrix, optionally ignoring background touches.
    cle.count_touching_neighbors(input_image: ndarray, ignore_background: bool = True) -> ndarray
    
    * Generate an image with pixel values of 1 on label edges and 0 elsewhere from an input labelmap.
    cle.detect_label_edges(input_image: ndarray) -> ndarray
    
    * Perform a Difference of Gaussian operation on a 3D image by applying two Gaussian blurs with specified sigmas and subtracting the results, suitable for 32-bit Float images.
    cle.difference_of_gaussian(input_image: ndarray, sigma1_x: float = 2, sigma1_y: float = 2, sigma1_z: float = 2, sigma2_x: float = 2, sigma2_y: float = 2, sigma2_z: float = 2) -> ndarray
    
    * Generate a binary image using dilation based on the specified 'box' or 'sphere' connectivity from an input image, with non-zero pixels treated as ones.
    cle.dilate(input_image: ndarray, connectivity: str = 'box') -> ndarray
    
    * Enlarge labels in an isotropic input image without overlap, optionally specifying output image, dilation radius, and device.
    cle.dilate_labels(input_image: ndarray, radius: int = 2) -> ndarray
    
    * Generate a binary image with pixel values 0 and 1 by applying binary dilation on the input image considering the von Neumann neighborhood.
    cle.dilate_sphere(input_image: ndarray) -> ndarray
    
    * apply Gaussian blur to an image, divide the original by the result, and optionally choose sigma values and device settings
    cle.divide_by_gaussian_background(input_image: ndarray, sigma_x: float = 2, sigma_y: float = 2, sigma_z: float = 2) -> ndarray
    
    * perform binary erosion on an input image using either 'box' or 'sphere' connectivity, with pixel interpretation for non-zero values as 1.
    cle.erode(input_image: ndarray, connectivity: str = 'box') -> ndarray
    
    * Perform binary erosion on an input image using the von Neumann neighborhood, outputting a binary image with pixel values 0 and 1.
    cle.erode_sphere(input_image: ndarray) -> ndarray
    
    * segment and label an image using blurring, thresholding, erosion, and masked Voronoi-labeling, suitable for dense objects but may remove small objects.
    cle.eroded_otsu_labeling(input_image: ndarray, number_of_erosions: int = 5, outline_sigma: float = 2) -> ndarray
    
    * Remove edge-touching labels from an image and renumber remaining labels, with options to exclude edges along x, y, z axes and specify a computing device.
    cle.exclude_labels_on_edges(input_image: ndarray, exclude_x: bool = True, exclude_y: bool = True, exclude_z: bool = True) -> ndarray
    
    * Filter a label image to remove labels smaller than a specified minimum size, optionally specifying an output image and processing device.
    cle.exclude_large_labels(input_image: ndarray, min_size: float = 100) -> ndarray
    
    * Filter out labels in an image that exceed a specified maximum size.
    cle.exclude_small_labels(input_image: ndarray, max_size: float = 100) -> ndarray
    
    * Dilate regions in a label map image using an octagon shape until they touch, with results output to a specified device or image.
    cle.extend_labeling_via_voronoi(input_image: ndarray) -> ndarray
    
    * Label objects in grey-value images using Gaussian blur, Otsu-thresholding, and connected component labeling, with adjustable segmentation via outline_sigma.
    cle.gauss_otsu_labeling(input_image0: ndarray, outline_sigma: float = 0) -> ndarray
    
    * Apply a Gaussian blur to an image with specified sigma values for the X, Y, and Z axes, allowing for non-isotropic filtering and skipping blur in any direction with a zero sigma.
    cle.gaussian_blur(input_image: ndarray, sigma_x: float = 0, sigma_y: float = 0, sigma_z: float = 0) -> ndarray
    
    * Calculate a matrix representing the distances between points from two n-dimensional pointlists extracted from input images, with the resulting matrix size being (n+1) x (m+1).
    cle.generate_distance_matrix(input_image0: ndarray, input_image1: ndarray) -> ndarray
    
    * generate a matrix representing label adjacency by marking touching labels in an input image, while utilizing an optional output image and device parameter.
    cle.generate_touch_matrix(input_image: ndarray) -> ndarray
    
    * Convert a binary image with single pixels set to 1 into a labeled spots image, where each detected spot is assigned a unique number.
    cle.label_spots(input_image: ndarray) -> ndarray
    
    * Generate a list of coordinates for labelled points in an image based on a labelmap resulting from connected components analysis.
    cle.labelled_spots_to_pointlist(input_image: ndarray) -> ndarray
    
    * Apply the Laplace operator using 'box' or 'sphere' connectivity to process an input image, optionally specifying an output image and device.
    cle.laplace(input_image: ndarray, connectivity: str = 'box') -> ndarray
    
    * Apply the Laplace operator using a Box neighborhood to process an input image, with options to specify an output image and the device for computation.
    cle.laplace_box(input_image: ndarray) -> ndarray
    
    * Apply the Laplace operator using a Diamond neighborhood to process an input image, optionally specifying an output image and device.
    cle.laplace_diamond(input_image: ndarray) -> ndarray
    
    * Apply a binary mask to an image, copying pixels from the input image to the output where the mask is non-zero, resulting in a masked image.
    cle.mask(input_image: ndarray, mask: ndarray) -> ndarray
    
    * Generate a masked image by copying pixels from an intensity image where corresponding pixels in a label image match a specified label value, with optional output image and device parameters.
    cle.mask_label(input_image0: ndarray, input_image1: ndarray, label: float = 1) -> ndarray
    
    * Label connected components and dilate them using an octagon shape within a masked area to produce an output label map, with optional device specification.
    cle.masked_voronoi_labeling(input_image: ndarray, mask: ndarray) -> ndarray
    
    * Compare each pixel value in an image with a constant scalar and output the maximum between them.
    cle.maximum_image_and_scalar(input_image: ndarray, scalar: float = 0) -> ndarray
    
    * Determine the maximum pixel value between two images, optionally storing the result in a specified output and selecting the device for computation.
    cle.maximum_images(input_image0: ndarray, input_image1: ndarray) -> ndarray
    
    * Calculate the local maximum within a spherical neighborhood for each pixel in an image, with customizable radii along the x, y, and z axes, and optionally specify an output image and device for processing.
    cle.maximum_sphere(input_image: ndarray, radius_x: float = 1, radius_y: float = 1, radius_z: float = 0) -> ndarray
    
    * calculate the maximum intensity projection of an image along the Z-axis, optionally specifying an output image and a processing device.
    cle.maximum_z_projection(input_image: ndarray) -> ndarray
    
    * Calculate the local mean average of pixels within a specified neighborhood shape (box or sphere) in an image, with adjustable radius for each dimension.
    cle.mean(input_image: ndarray, radius_x: int = 1, radius_y: int = 1, radius_z: int = 1, connectivity: str = 'box') -> ndarray
    
    * Calculate the local mean average of pixels within a spherical neighborhood defined by specified radii in an image.
    cle.mean_sphere(input_image: ndarray, radius_x: int = 1, radius_y: int = 1, radius_z: int = 1) -> ndarray
    
    * Calculate and store the mean squared error (MSE) between two images in the ImageJs Results table.
    cle.mean_squared_error(input_image0: ndarray, input_image1: ndarray) -> float
    
    * compute the mean average intensity projection of an image along the Z-axis, with optional output image and device specifications
    cle.mean_z_projection(input_image: ndarray) -> ndarray
    
    * Determine the local minimum in a spherical neighborhood of each pixel with specified radii in the x, y, and z axes on a selected device.
    cle.minimum_sphere(input_image: ndarray, radius_x: float = 1, radius_y: float = 1, radius_z: float = 1) -> ndarray
    
    * Compute the minimum intensity projection of an image along the Z-axis, specifying optional output and device parameters.
    cle.minimum_z_projection(input_image: ndarray) -> ndarray
    
    * compute the local mode of a pixel's neighborhood in an image with specified radius and shape, using values from 0 to 255.
    cle.mode(input_image: ndarray, radius_x: int = 1, radius_y: int = 1, radius_z: int = 1, connectivity: str = 'box') -> ndarray
    
    * calculate the local mode of pixel intensities within a spherical neighborhood for semantic segmentation correction, with intensity values ranging from 0 to 255, using specified radii and device options.
    cle.mode_sphere(input_image: ndarray, radius_x: int = 1, radius_y: int = 1, radius_z: int = 1) -> ndarray
    
    * Compute the remainder of division for corresponding pixel values between two images, optionally specifying an output image and device.
    cle.modulo_images(input_image0: ndarray, input_image1: ndarray) -> ndarray
    
    * Apply morphological opening to intensity or binary images using a spherical or box-shaped footprint with configurable radii and device options.
    cle.opening(input_image: ndarray, radius_x: float = 1, radius_y: float = 1, radius_z: float = 0, connectivity: str = 'box') -> ndarray
    
    * Perform a morphological opening on a label image using an octagonal structuring element, with optional output image and device specification.
    cle.opening_labels(input_image: ndarray, radius: int = 0) -> ndarray
    
    * Apply morphological opening to intensity or binary images using a spherical footprint, specifying radii along x, y, and z axes, with optional output and device parameters.
    cle.opening_sphere(input_image: ndarray, radius_x: float = 1, radius_y: float = 1, radius_z: float = 0) -> ndarray
    
    * extract pixel intensities from specified (x,y,z) coordinates in an image and store them in a vector
    cle.read_values_from_positions(input_image: ndarray, list: ndarray) -> ndarray
    
    * Reduce a label map to its centroids in an optional output image using a specified device.
    cle.reduce_labels_to_centroids(input_image: ndarray) -> ndarray
    
    * Extracts label edges from a label map, preserving label IDs and setting the background to zero, with optional device specification for processing.
    cle.reduce_labels_to_label_edges(input_image: ndarray) -> ndarray
    
    * Renumber labels in an image to eliminate gaps, ensuring the number of labels equals the maximum label index, primarily processed on the CPU.
    cle.relabel_sequential(input_image: ndarray, blocksize: int = 4096) -> ndarray
    
    * Remove labels from the edges of an image and renumber remaining elements, with options to exclude specific axes and specify a device for operation.
    cle.remove_labels_on_edges(input_image: ndarray, exclude_x: bool = True, exclude_y: bool = True, exclude_z: bool = True) -> ndarray
    
    * Filter out objects larger than a specified size from a label map image.
    cle.remove_large_labels(input_image: ndarray, max_size: float = 100) -> ndarray
    
    * Remove labelled objects smaller than a specified pixel size from a label map image, with optional output image specification and device selection.
    cle.remove_small_labels(input_image: ndarray, min_size: float = 100) -> ndarray
    
    * Replace the intensities of an input image based on a vector mapping old to new intensity values, with optional specification of the output image and device for processing.
    cle.replace_intensities(input_image0: ndarray, input_image1: ndarray) -> ndarray
    
    * Replace pixel intensities in an image using a vector where the index corresponds to old intensities and the value to new ones, with optional output image and device specification.
    cle.replace_values(input_image0: ndarray, input_image1: ndarray) -> ndarray
    
    * Translate and rotate an image by specified vectors and angles (in degrees) optionally with interpolation and resizing, supporting operation on specified devices.
    cle.rigid_transform(input_image: ndarray, translate_x: float = 0, translate_y: float = 0, translate_z: float = 0, angle_x: float = 0, angle_y: float = 0, angle_z: float = 0, centered: bool = True, interpolate: bool = False, resize: bool = False) -> ndarray
    
    * Perform a morphological opening on a label image, fill gaps using Voronoi labeling, and retain the background, suitable for isotropic images.
    cle.smooth_labels(input_image: ndarray, radius: int = 0) -> ndarray
    
    * Apply the Sobel kernel to an input image for edge detection with optional device specification and output.
    cle.sobel(input_image: ndarray) -> ndarray
    
    * Calculate the squared pixel-by-pixel difference between two images, optionally specifying an output image and a device for processing.
    cle.squared_difference(input_image0: ndarray, input_image1: ndarray) -> ndarray
    
    * Calculate the local standard deviation of an image's pixel neighborhood with specified radii and optional output image, using either a box or sphere connectivity shape.
    cle.standard_deviation(input_image: ndarray, radius_x: int = 1, radius_y: int = 1, radius_z: int = 1, connectivity: str = 'box') -> ndarray
    
    * Calculate the local standard deviation in a pixel's spherical neighborhood based on specified radii along the x, y, and (optionally) z axes for either 2D or 3D images, with an optional device specification for processing.
    cle.standard_deviation_sphere(input_image: ndarray, radius_x: int = 1, radius_y: int = 1, radius_z: int = 1) -> ndarray
    
    * compute the standard deviation intensity projection of an image stack along the Z-axis, with optional specification of output image and device.
    cle.std_z_projection(input_image: ndarray) -> ndarray
    
    * Subtracts a Gaussian-blurred version of an image from the original, with configurable blur radii and optional output image and device specification.
    cle.subtract_gaussian_background(input_image: ndarray, sigma_x: float = 2, sigma_y: float = 2, sigma_z: float = 2) -> ndarray
    
    * Calculate the sum intensity projection of an image along the Z-axis, optionally specifying an output image and device.
    cle.sum_z_projection(input_image: ndarray) -> ndarray
    
    * Convert an image to binary using Otsu's thresholding via a GPU-accelerated histogram.
    cle.threshold_otsu(input_image: ndarray) -> ndarray
    
    * perform background subtraction on an image using a tophat filter with customizable radii and connectivity options
    cle.top_hat(input_image: ndarray, radius_x: float = 1, radius_y: float = 1, radius_z: float = 1, connectivity: str = 'box') -> ndarray
    
    * Subtract the background from an input image using a tophat filter with customizable radii and device options.
    cle.top_hat_sphere(input_image: ndarray, radius_x: float = 1, radius_y: float = 1, radius_z: float = 1) -> ndarray
    
    * Calculate the local variance of a pixel's spherical neighborhood in an image, considering specified radii for each axis and optionally designating output and processing device.
    cle.variance_sphere(input_image: ndarray, radius_x: int = 1, radius_y: int = 1, radius_z: int = 1) -> ndarray
    
    * Label connected components in a binary image and dilate regions using an octagon shape until they touch, returning the resulting label map.
    cle.voronoi_labeling(input_image: ndarray) -> ndarray
    
    * Label objects in isotropic greyvalue images by applying Gaussian blurs, spot detection, Otsu thresholding, and Voronoi labeling with parameters for segmentation precision and cell proximity control.
    cle.voronoi_otsu_labeling(input_image: ndarray, spot_sigma: float = 2, outline_sigma: float = 2) -> ndarray
    
    """



import numpy as np
import pytest

import pyclesperanto as cle

cle.select_device("TX")


# @pytest.mark.parametrize("shape", 2 ** np.arange(9, 11), ids=lambda x: f"{x}x{x}")
# @pytest.mark.parametrize("target", ["cpu", "gpu"])
# def test_multiply_matrix(shape, benchmark, target):
#     @benchmark
#     def multiply():
#         a_np = np.random.rand(shape, shape).astype("float32")
#         b_np = np.random.rand(shape, shape).astype("float32")
#         if target == "gpu":
#             # push arrays to GPU
#             gpu_a = cle.push(a_np)
#             gpu_b = cle.push(b_np)
#             # allocate memory for result on GPU
#             gpu_c = cle.create((shape, shape), dtype=float)
#             cle.multiply_matrix(gpu_a, gpu_b, gpu_c)
#             _ = gpu_c.get()
#         else:
#             # multiply matrix on CPU
#             _ = np.dot(a_np.T, b_np.T)
#         # np.testing.assert_allclose(cle.pull(gpu_c), cpu_c, atol=1e-3)


# make a test for matrix multiplication
def test_multiply_matrix():
    matrix_a = np.random.rand(3, 3).astype("float32")
    matrix_b = np.random.rand(3, 3).astype("float32")

    valid = np.dot(matrix_a, matrix_b)
    matrix_c = cle.multiply_matrix(matrix_a, matrix_b)

    result = cle.pull(matrix_c)

    print(result)
    print(valid)

    assert np.allclose(result, valid, atol=0.001)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


# this test may only work on a machine with multiple GPUs
def test_gpu_switch():
    image = np.random.random((100, 100))

    # select NVidia
    print(cle.select_device("RTX"))
    cle.gaussian_blur(image, sigma_x=1, sigma_y=1)

    # select AMD
    print(cle.select_device("gfx"))
    cle.gaussian_blur(image, sigma_x=1, sigma_y=1)

    # select Intel
    print(cle.select_device("Intel"))
    cle.gaussian_blur(image, sigma_x=1, sigma_y=1)



#
# This code is auto-generated from CLIc 'cle::tier4.hpp' file, do not edit manually.
#

import importlib
import warnings
from typing import Optional

import numpy as np

from ._array import Image
from ._core import Device
from ._decorators import plugin_function

clic = importlib.import_module('._pyclesperanto', package='pyclesperanto')

@plugin_function
def label_bounding_box(
    input_image: Image,
    label_id: int,
    device: Optional[Device] =None
) -> list:
    """Determines the bounding box of the specified label from a label image. The
    positions are returned in  an array of 6 values as follows: minX, minY, minZ,
    maxX, maxY, maxZ.

    Parameters
    ----------
    input_image: Image 
        
    label_id: int 
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    list

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_boundingBox
    """
    return clic._label_bounding_box(device, input_image, int(label_id))

@plugin_function(categories=["in assistant", "combine", "bia-bob-suggestion"])
def mean_squared_error(
    input_image0: Image,
    input_image1: Image,
    device: Optional[Device] =None
) -> float:
    """Determines the mean squared error (MSE) between two images. The MSE will be
    stored in a new row of ImageJs Results table in the column 'MSE'.

    Parameters
    ----------
    input_image0: Image 
        
    input_image1: Image 
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    float

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_meanSquaredError
    """
    return clic._mean_squared_error(device, input_image0, input_image1)

@plugin_function
def spots_to_pointlist(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Transforms a spots image as resulting from maximum/minimum detection in an image
    where every column contains d pixels (with d = dimensionality of the original
    image) with the coordinates of the maxima/minima.

    Parameters
    ----------
    input_image: Image 
        
    output_image: Optional[Image] (= None)
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_spotsToPointList
    """
    return clic._spots_to_pointlist(device, input_image, output_image)

@plugin_function(categories=["label processing", "in assistant", "bia-bob-suggestion"])
def relabel_sequential(
    input_image: Image,
    output_image: Optional[Image] =None,
    blocksize: int =4096,
    device: Optional[Device] =None
) -> Image:
    """Analyses a label map and if there are gaps in the indexing (e.g. label 5 is not
    present) all subsequent labels will be relabelled. Thus, afterwards number of
    labels and maximum label index are equal. This operation is mostly performed on
    the CPU.

    Parameters
    ----------
    input_image: Image 
        Label image.
    output_image: Optional[Image] (= None)
        Output label image.
    blocksize: int (= 4096)
        Renumbering is done in blocks for performance reasons.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_closeIndexGapsInLabelMap
    """
    return clic._relabel_sequential(device, input_image, output_image, int(blocksize))

@plugin_function(categories=["binarize", "in assistant", "bia-bob-suggestion"])
def threshold_otsu(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Binarizes an image using Otsu's threshold method [3] implemented in
    scikit-image[2] using a histogram determined on the GPU to create binary images.

    Parameters
    ----------
    input_image: Image 
        Input image to threshold.
    output_image: Optional[Image] (= None)
        Output binary image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_thresholdOtsu
    [2] https://scikit-image.org/docs/dev/api/skimage.filters.html#skimage.filters.threshold_otsu
    [3] https://ieeexplore.ieee.org/document/4310076
    """
    return clic._threshold_otsu(device, input_image, output_image)

@plugin_function(categories=["label measurement", "map", "in assistant"])
def pixel_count_map(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Takes a label map, determines the number of pixels per label and replaces every
    label with the that number. This results in a parametric image expressing area
    or volume.

    Parameters
    ----------
    input_image: Image 
        Label image to measure
    output_image: Optional[Image] (= None)
        Parametric image computed
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_pixelCountMap
    """
    return clic._pixel_count_map(device, input_image, output_image)

@plugin_function(categories=["label measurement", "map", "in assistant"])
def label_pixel_count_map(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Takes a label map, determines the number of pixels per label and replaces every
    label with the that number. This results in a parametric image expressing area
    or volume.

    Parameters
    ----------
    input_image: Image 
        Label image to measure
    output_image: Optional[Image] (= None)
        Parametric image computed
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_pixelCountMap
    """
    return clic._label_pixel_count_map(device, input_image, output_image)

@plugin_function
def centroids_of_labels(
    input_image: Image,
    output_image: Optional[Image] =None,
    withBG: bool =False,
    device: Optional[Device] =None
) -> Image:
    """Determines the centroids of all labels in a label image or image stack. It
    writes the resulting coordinates in point list image of dimensions n * d where n
    is the number of labels and d=3 the dimensionality (x,y,z) of the original
    image.

    Parameters
    ----------
    input_image: Image 
        Label image where the centroids will be determined from.
    output_image: Optional[Image] (= None)
        Output image where the centroids will be written to.
    withBG: bool (= False)
        Determines if the background label should be included.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_centroidsOfLabels
    """
    return clic._centroids_of_labels(device, input_image, output_image, withBG)

@plugin_function(categories=["label processing", "combine"])
def remove_labels_with_map_values_out_of_range(
    input_image: Image,
    values: Image,
    output_image: Optional[Image] =None,
    min_value: float =0,
    max_value: float =100,
    device: Optional[Device] =None
) -> Image:
    """Remove labels with values outside a given value range based on a vector of
    values associated with the labels.

    Parameters
    ----------
    input_image: Image 
        Input image where labels will be filtered.
    values: Image 
        Vector of
    output_image: Optional[Image] (= None)
        Output image where labels will be written to.
    min_value: float (= 0)
        Minimum value to keep.
    max_value: float (= 100)
        Maximum value to keep.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_excludeLabelsWithValuesOutOfRange
    """
    return clic._remove_labels_with_map_values_out_of_range(device, input_image, values, output_image, float(min_value), float(max_value))

@plugin_function(categories=["label processing", "combine"])
def remove_labels_with_map_values_within_range(
    input_image: Image,
    values: Image,
    output_image: Optional[Image] =None,
    min_value: float =0,
    max_value: float =100,
    device: Optional[Device] =None
) -> Image:
    """Remove labels with values inside a given value range based on a vector of values
    associated with the labels.

    Parameters
    ----------
    input_image: Image 
        Input image where labels will be filtered.
    values: Image 
        Vector of
    output_image: Optional[Image] (= None)
        Output image where labels will be written to.
    min_value: float (= 0)
        Minimum value to keep.
    max_value: float (= 100)
        Maximum value to keep.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_excludeLabelsWithValuesWithinRange
    """
    return clic._remove_labels_with_map_values_within_range(device, input_image, values, output_image, float(min_value), float(max_value))

@plugin_function(categories=["label processing", "combine"])
def exclude_labels_with_map_values_out_of_range(
    input_image: Image,
    values: Image,
    output_image: Optional[Image] =None,
    min_value_range: float =0,
    max_value_range: float =100,
    device: Optional[Device] =None
) -> Image:
    """Exclude labels with values outside a given value range based on a vector of
    values associated with the labels.

    Parameters
    ----------
    input_image: Image 
        Input image where labels will be filtered.
    values: Image 
        Vector of
    output_image: Optional[Image] (= None)
        Output image where labels will be written to.
    min_value_range: float (= 0)
        Minimum value to keep.
    max_value_range: float (= 100)
        Maximum value to keep.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_excludeLabelsWithValuesOutOfRange
    """
    return clic._exclude_labels_with_map_values_out_of_range(device, input_image, values, output_image, float(min_value_range), float(max_value_range))

@plugin_function(categories=["label processing", "combine"])
def exclude_labels_with_map_values_within_range(
    input_image: Image,
    values: Image,
    output_image: Optional[Image] =None,
    min_value_range: float =0,
    max_value_range: float =100,
    device: Optional[Device] =None
) -> Image:
    """Exclude labels with values inside a given value range based on a vector of
    values associated with the labels.

    Parameters
    ----------
    input_image: Image 
        Input image where labels will be filtered.
    values: Image 
        Vector of
    output_image: Optional[Image] (= None)
        Output image where labels will be written to.
    min_value_range: float (= 0)
        Minimum value to keep.
    max_value_range: float (= 100)
        Maximum value to keep.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_excludeLabelsWithValuesWithinRange
    """
    return clic._exclude_labels_with_map_values_within_range(device, input_image, values, output_image, float(min_value_range), float(max_value_range))

@plugin_function(categories=["label processing", "in assistant", "map"])
def extension_ratio_map(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines the ratio of the extension for every label in a label map and returns
    it as a parametric map. The extension ration is defined as the maximum distance
    of any pixel in the label to the label's centroid divided by the average
    distance of all pixels in the label to the centroid.

    Parameters
    ----------
    input_image: Image 
        Input label image.
    output_image: Optional[Image] (= None)
        Output parametric image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_extensionRatioMap
    """
    return clic._extension_ratio_map(device, input_image, output_image)


from typing import Optional, Union

import numpy as np

from . import _operators
from ._core import Device, get_device
from ._pyclesperanto import _Array as Array
from ._utils import _assert_supported_dtype


def _prepare_array(arr) -> np.ndarray:
    """Converts a given array to a numpy array with C memory layout.

    Parameters
    ----------
    arr :
        The array to convert.

    Returns
    -------
    np.ndarray
        The converted array.
    """
    return np.require(arr, None, "C")


def __str__(self) -> str:
    """Returns a string representation of the Array."""
    return self.get().__str__()


def __repr__(self) -> str:
    """Returns a string representation of the Array."""
    repr_str = self.get().__repr__()
    extra_info = f"mtype={self.mtype}"
    return repr_str[:-1] + f", {extra_info})"


def set(
    self,
    array: np.ndarray,
    origin: Optional[tuple] = None,
    region: Optional[tuple] = None,
) -> None:
    """Set the content of the Array to the given numpy array.

    Parameters
    ----------
    array : np.ndarray
        The array to set from.
    origin : tuple, optional
        The origin of the region of interest, by default None
    region : tuple, optional
        The region of interest, by default None

    Returns
    -------
    Array
        The array itself.
    """
    if not isinstance(array, (np.ndarray, Array)):
        array = np.array(array)

    if array.dtype != self.dtype:
        array = array.astype(self.dtype)

    if region and array.size != np.prod(region):
        raise IndexError(
            f"Value size mismatch the targeted region: {array.size} != {np.prod(region)} ({array.shape} != {tuple(np.squeeze(region))})"
        )
    elif not region and self.size != array.size:
        raise IndexError(
            f"Value size mismatch the targeted region: {self.size} != {array.size} ({self.shape} != {array.shape})"
        )

    self._write(_prepare_array(array), origin, region)
    return self


def get(
    self, origin: Optional[tuple] = None, region: Optional[tuple] = None
) -> np.ndarray:
    """Get the content of the Array into a numpy array.

    Parameters
    ----------
    origin : tuple, optional
        The origin of the region of interest, by default None
    region : tuple, optional
        The region of interest, by default None

    Returns
    -------
    np.ndarray
        The array itself.
    """
    caster = {
        "float32": self._read_float32,
        "int8": self._read_int8,
        "int16": self._read_int16,
        "int32": self._read_int32,
        # "int64": self._read_int64,
        "uint8": self._read_uint8,
        "uint16": self._read_uint16,
        "uint32": self._read_uint32,
        # "uint64": self._read_uint64,
    }
    return caster[self.dtype.name](origin, region)


def __array__(self, dtype=None) -> np.ndarray:
    """Returns a numpy array representation of the Array."""
    if dtype is None:
        return self.get()
    else:
        return self.get().astype(dtype)


def to_device(cls, arr, *args, **kwargs):
    """Create an Array object from a numpy array (same shape, dtype, and memory).

    Parameters
    ----------
    arr : np.ndarray
        The array to convert.
    mtype : str, optional
        The memory type, by default "buffer"
    device : Device, optional
        The device, by default None

    Returns
    -------
    Array
        The converted array.
    """
    if isinstance(arr, Array):
        return arr
    mtype = kwargs.get("mtype", "buffer")
    device = kwargs.get("device", get_device())
    return cls.create(arr.shape, arr.dtype, mtype, device).set(arr)


def from_array(cls, arr, *args, **kwargs):
    """Create an Array object from a numpy array (same shape, dtype, and memory).

    Parameters
    ----------
    arr : np.ndarray
        The array to convert.
    mtype : str, optional
        The memory type, by default "buffer"
    device : Device, optional
        The device, by default None

    Returns
    -------
    Array
        The converted array.
    """
    _assert_supported_dtype(arr.dtype)
    return cls.to_device(arr, *args, **kwargs)


def empty(cls, shape, dtype=float, *args, **kwargs):
    """Create an empty Array object from a shape.

    Parameters
    ----------
    shape : tuple, list or np.ndarray
        The shape of the array, maximum 3 elements.
    dtype : np.dtype, default float
        The dtype of the array.
    mtype : str, optional
        The memory type, by default "buffer"
    device : Device, optional
        The device, by default None

    Returns
    -------
    Array
        A new Array object.
    """
    _assert_supported_dtype(dtype)
    mtype = kwargs.get("mtype", "buffer")
    device = kwargs.get("device", get_device())

    return cls.create(shape=shape, dtype=dtype, mtype=mtype, device=device)


def empty_like(cls, arr):
    """Create an empty Array object from an other array.

    Parameters
    ----------
    arr : np.ndarray or Array or other array-like structure
        The array to create like.

    Returns
    -------
    Array
        The created array.
    """
    _assert_supported_dtype(arr.dtype)
    mtype = arr.mtype if isinstance(arr, Array) else "buffer"
    device = arr.device if isinstance(arr, Array) else get_device()
    return Array.create(arr.shape, arr.dtype, mtype, device)


def zeros(cls, shape, dtype=float, *args, **kwargs):
    """Create an Array object full of zeros from a shape.

    Parameters
    ----------
    shape : tuple, list or np.ndarray
        The shape of the array, maximum 3 elements.
    dtype : np.dtype, default float
        The dtype of the array.
    mtype : str, optional
        The memory type, by default "buffer"
    device : Device, optional
        The device, by default None

    Returns
    -------
    Array
        The created array.
    """
    _assert_supported_dtype(dtype)
    new_array = cls.empty(shape=shape, dtype=dtype, *args, **kwargs)
    new_array.fill(0)
    return new_array


def zeros_like(cls, arr):
    """Create an Array object filled with zeros from an other array.

    Parameters
    ----------
    arr : np.ndarray or Array or other array-like structure
        The array to create like.

    Returns
    -------
    Array
        The created array.
    """
    _assert_supported_dtype(arr.dtype)
    mtype = arr.mtype if isinstance(arr, Array) else "buffer"
    device = arr.device if isinstance(arr, Array) else get_device()
    return cls.zeros(shape=arr.shape, dtype=arr.dtype, mtype=mtype, device=device)


def T(self):
    """Transpose the Array. Only works for 2D and 3D arrays."""
    from ._tier1 import transpose_xy, transpose_xz

    if len(self.shape) == 2:
        return transpose_xy(self)
    elif len(self.shape) == 3:
        return transpose_xz(self)
    else:
        return self


def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
    if method == "__call__":
        func = getattr(Array, f"__{ufunc.__name__}__", None)
        if func is not None:
            return func(*[Array.to_device(i) for i in inputs], **kwargs)
    return NotImplemented


def reshape(self, shape):
    """Reshape the Array."""
    return self.get().reshape(shape)


# missing operators:
# __array_interface__
# __array_ufunc__

# Add class methods, properties and magic methods
setattr(Array, "T", property(T))
setattr(Array, "set", set)
setattr(Array, "get", get)
setattr(Array, "__array_ufunc__", __array_ufunc__)
setattr(Array, "__str__", __str__)
setattr(Array, "__repr__", __repr__)
setattr(Array, "__array__", __array__)
setattr(Array, "from_array", classmethod(from_array))
setattr(Array, "empty", classmethod(empty))
setattr(Array, "empty_like", classmethod(empty_like))
setattr(Array, "zeros", classmethod(zeros))
setattr(Array, "zeros_like", classmethod(zeros_like))
setattr(Array, "to_device", classmethod(to_device))
setattr(Array, "reshape", reshape)

# Add operations and class methods from _operators module
setattr(Array, "astype", _operators._astype)
setattr(Array, "max", _operators._max)
setattr(Array, "min", _operators._min)
setattr(Array, "sum", _operators._sum)
setattr(Array, "__pos__", _operators.__pos__)
setattr(Array, "__neg__", _operators.__neg__)
setattr(Array, "__add__", _operators.__add__)
setattr(Array, "__iadd__", _operators.__iadd__)
setattr(Array, "__sub__", _operators.__sub__)
setattr(Array, "__div__", _operators.__div__)
setattr(Array, "__truediv__", _operators.__truediv__)
setattr(Array, "__idiv__", _operators.__idiv__)
setattr(Array, "__itruediv__", _operators.__itruediv__)
setattr(Array, "__mul__", _operators.__mul__)
setattr(Array, "__imul__", _operators.__imul__)
setattr(Array, "__gt__", _operators.__gt__)
setattr(Array, "__ge__", _operators.__ge__)
setattr(Array, "__lt__", _operators.__lt__)
setattr(Array, "__le__", _operators.__le__)
setattr(Array, "__eq__", _operators.__eq__)
setattr(Array, "__ne__", _operators.__ne__)
setattr(Array, "__pow__", _operators.__pow__)
setattr(Array, "__ipow__", _operators.__ipow__)
setattr(Array, "_plt_to_png", _operators.__plt_to_png__)
setattr(Array, "_png_to_html", _operators.__png_to_html__)
setattr(Array, "_repr_html_", _operators.__repr_html__)
setattr(Array, "__iter__", _operators.__iter__)
setattr(Array, "__setitem__", _operators.__setitem__)
setattr(Array, "__getitem__", _operators.__getitem__)

# Create Image type
Image = Union[np.ndarray, Array]


def is_image(object):
    """Returns True if the given object is an image."""
    return (
        isinstance(object, np.ndarray)
        or isinstance(object, tuple)
        or isinstance(object, list)
        or isinstance(object, Array)
        or str(type(object))
        in [
            "<class 'cupy._core.core.ndarray'>",
            "<class 'dask.array.core.Array'>",
            "<class 'xarray.core.dataarray.DataArray'>",
            "<class 'resource_backed_dask_array.ResourceBackedDaskArray'>",
            "<class 'torch.Tensor'>",
            "<class 'pyclesperanto_prototype._tier0._pycl.OCLArray'>"
        ]
    )



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_crop_border_2d():
    test1 = cle.push(
        np.asarray([[0, 0, 0, 1], [0, 0, 3, 1], [0, 0, 3, 1], [1, 1, 1, 1]])
    )

    reference = cle.push(np.asarray([[0, 3], [0, 3]]))

    result = cle.crop_border(test1, border_size=1)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)
    assert np.array_equal(a, b)


def test_crop_3d():
    input_image = cle.push(
        np.asarray(
            [
                [[0, 0, 0, 1], [0, 0, 3, 1], [0, 0, 3, 1], [1, 1, 1, 1]],
                [[0, 0, 0, 1], [0, 1, 3, 1], [0, 0, 3, 1], [1, 1, 1, 1]],
                [[0, 0, 0, 1], [0, 0, 3, 1], [0, 0, 3, 1], [1, 1, 1, 1]],
                [[0, 0, 0, 1], [0, 0, 3, 1], [0, 0, 3, 1], [1, 1, 1, 1]],
            ]
        )
    )
    reference = cle.push(
        np.asarray(
            [
                [[1, 3], [0, 3]],
                [[0, 3], [0, 3]],
            ]
        )
    )

    result = cle.crop_border(input_image, border_size=1)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)
    assert np.array_equal(a, b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_radians_to_degrees():
    test = cle.push(np.asarray([[np.pi, 0, -0.5 * np.pi]]))

    reference = cle.push(np.asarray([[180, 0, -90]]))

    result = cle.radians_to_degrees(test)

    a = cle.pull(result)
    b = cle.pull(reference)

    assert np.allclose(a, b, 0.01)



import inspect
from functools import wraps
from typing import Callable, Optional, get_args

from toolz import curry  # type: ignore

from ._array import Array, Image, is_image
from ._core import Device, get_device
from ._memory import push


@curry
def plugin_function(
    function: Callable,
    categories: Optional[list] = None,
) -> Callable:
    """Function decorator to ensure correct types and values of all parameters.

    The given input parameters are either of type OpenCL data/image/buffer (which the GPU
    understands) or are converted to this type (see push function). If output
    parameters of type Image are not set, an empty image is created and
    handed over.

    Parameters
    ----------
    function : callable
        The function to be executed on the GPU.
    output_creator : callable, optional
        A function to create an output cleImage given an input cleImage. By
        default, we create output images of the same shape and type as input
        images.
    device_selector : callable, optional
        A function to select a device. By default, we use the current device instance.
    categories : list of str, optional
        A list of category names the function is associated with

    Returns
    -------
    worker_function : callable
        The actual function call that will be executed, magically creating
        output arguments of the correct type.
    """

    # function.fullargspec = inspect.getfullargspec(function)
    function.categories = categories
    # function.priority = priority

    @wraps(function)
    def worker_function(*args, **kwargs):
        sig = inspect.signature(function)
        bound = sig.bind(*args, **kwargs)
        bound.apply_defaults()

        # Get the device to use from the arguments or the input image or the current device
        input_device = next(
            (value for value in bound.arguments.values() if isinstance(value, Device)),
            None,
        )
        input_image = next(
            (value for value in bound.arguments.values() if is_image(value)), None
        )
        input_image_device = (
            getattr(input_image, "device", None) if input_image is not None else None
        )

        # if device is `cpu` or a string, set it to None
        if input_image_device == "cpu":
            input_image_device = None

        # Use input_device if available, else use the device of the input_image if it has the attribute device, else None
        use_device = input_device or input_image_device or get_device()

        # loop on all argument and if it is an image, push it to the device
        # if it is a device, set it to the use_device
        # if it is any other argument, do nothing
        for key, value in bound.arguments.items():
            if (
                is_image(value)
                and key in sig.parameters
                and (
                    sig.parameters[key].annotation is Image
                    or Array in get_args(sig.parameters[key].annotation)
                )
            ):
                bound.arguments[key] = push(value, device=use_device)
            if (
                key in sig.parameters
                and (
                    sig.parameters[key].annotation is Device
                    or Device in get_args(sig.parameters[key].annotation)
                )
                and value is None
            ):
                bound.arguments[key] = use_device

        # call the decorated function
        result = function(*bound.args, **bound.kwargs)

        return result

    worker_function.__module__ = "pyclesperanto"

    return worker_function



#
# This code is auto-generated from CLIc 'cle::tier2.hpp' file, do not edit manually.
#

import importlib
import warnings
from typing import Optional

import numpy as np

from ._array import Image
from ._core import Device
from ._decorators import plugin_function

clic = importlib.import_module('._pyclesperanto', package='pyclesperanto')

@plugin_function(categories=["combine", "in assistant"])
def absolute_difference(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines the absolute difference pixel by pixel between two images. <pre>f(x,
    y) = |x y| </pre>

    Parameters
    ----------
    input_image0: Image 
        The input image to be subtracted from.
    input_image1: Image 
        The input image which is subtracted.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_absoluteDifference
    """
    return clic._absolute_difference(device, input_image0, input_image1, output_image)

@plugin_function(categories=["combine", "in assistant"])
def add_images(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Calculates the sum of pairs of pixels x and y of two images X and Y. <pre>f(x,
    y) = x + y</pre>

    Parameters
    ----------
    input_image0: Image 
        The first input image to added.
    input_image1: Image 
        The second image to be added.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_addImages
    """
    return clic._add_images(device, input_image0, input_image1, output_image)

@plugin_function(categories=["filter", "background removal", "in assistant"])
def bottom_hat_box(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =1,
    radius_y: int =1,
    radius_z: int =1,
    device: Optional[Device] =None
) -> Image:
    """Apply a bottomhat filter for background subtraction to the input image.

    Parameters
    ----------
    input_image: Image 
        The input image where the background is subtracted from.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    radius_x: int (= 1)
        Radius of the background determination region in X.
    radius_y: int (= 1)
        Radius of the background determination region in Y.
    radius_z: int (= 1)
        Radius of the background determination region in Z.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_bottomHatBox
    """
    return clic._bottom_hat_box(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z))

@plugin_function(categories=["filter", "background removal", "in assistant"])
def bottom_hat_sphere(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: float =1,
    radius_y: float =1,
    radius_z: float =1,
    device: Optional[Device] =None
) -> Image:
    """Applies a bottomhat filter for background subtraction to the input image.

    Parameters
    ----------
    input_image: Image 
        The input image where the background is subtracted from.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    radius_x: float (= 1)
        Radius of the background determination region in X.
    radius_y: float (= 1)
        Radius of the background determination region in Y.
    radius_z: float (= 1)
        Radius of the background determination region in Z.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_bottomHatSphere
    """
    return clic._bottom_hat_sphere(device, input_image, output_image, float(radius_x), float(radius_y), float(radius_z))

@plugin_function(categories=["filter", "background removal", "in assistant"])
def bottom_hat(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: float =1,
    radius_y: float =1,
    radius_z: float =1,
    connectivity: str ="box",
    device: Optional[Device] =None
) -> Image:
    """Applies a bottomhat filter for background subtraction to the input image.

    Parameters
    ----------
    input_image: Image 
        The input image where the background is subtracted from.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    radius_x: float (= 1)
        Radius of the background determination region in X.
    radius_y: float (= 1)
        Radius of the background determination region in Y.
    radius_z: float (= 1)
        Radius of the background determination region in Z.
    connectivity: str (= "box")
        Element shape, "box" or "sphere"
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_bottomHatBox
    [2] https://clij.github.io/clij2-docs/reference_bottomHatSphere
    """
    return clic._bottom_hat(device, input_image, output_image, float(radius_x), float(radius_y), float(radius_z), str(connectivity))

@plugin_function(categories=["combine", "in assistant"])
def clip(
    input_image: Image,
    output_image: Optional[Image] =None,
    min_intensity: Optional[float] =None,
    max_intensity: Optional[float] =None,
    device: Optional[Device] =None
) -> Image:
    """Limits the range of values in an image. This function is supposed to work
    similarly as its counter part in numpy [1].

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    min_intensity: Optional[float] (= None)
        new, lower limit of the intensity range
    max_intensity: Optional[float] (= None)
        new, upper limit of the intensity range
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://numpy.org/doc/stable/reference/generated/numpy.clip.html
    """
    return clic._clip(device, input_image, output_image, min_intensity, max_intensity)

@plugin_function(categories=["filter", "in assistant"])
def closing_box(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =0,
    radius_y: int =0,
    radius_z: int =0,
    device: Optional[Device] =None
) -> Image:
    """Closing operator, boxshaped Applies morphological closing to intensity images
    using a boxshaped footprint. This operator also works with binary images.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 0)
        Radius along the x axis.
    radius_y: int (= 0)
        Radius along the y axis.
    radius_z: int (= 0)
        Radius along the z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._closing_box(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z))

@plugin_function(categories=["filter", "in assistant", "bia-bob-suggestion"])
def closing_sphere(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =1,
    radius_y: int =1,
    radius_z: int =0,
    device: Optional[Device] =None
) -> Image:
    """Closing operator, sphereshaped Applies morphological closing to intensity images
    using a sphereshaped footprint. This operator also works with binary images.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 1)
        Radius along the x axis.
    radius_y: int (= 1)
        Radius along the y axis.
    radius_z: int (= 0)
        Radius along the z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._closing_sphere(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z))

@plugin_function(categories=["filter", "in assistant"])
def closing(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =1,
    radius_y: int =1,
    radius_z: int =0,
    connectivity: str ="box",
    device: Optional[Device] =None
) -> Image:
    """Closing operator, sphereshaped Applies morphological closing to intensity images
    using a sphereshaped footprint. This operator also works with binary images.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 1)
        Radius along the x axis.
    radius_y: int (= 1)
        Radius along the y axis.
    radius_z: int (= 0)
        Radius along the z axis.
    connectivity: str (= "box")
        Element shape, "box" or "sphere"
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._closing(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z), str(connectivity))

@plugin_function(categories=["combine", "transform", "in assistant", "bia-bob-suggestion"])
def concatenate_along_x(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Concatenate two images or stacks along the X axis.

    Parameters
    ----------
    input_image0: Image 
        First input image.
    input_image1: Image 
        Second input image.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_combineHorizontally
    """
    return clic._concatenate_along_x(device, input_image0, input_image1, output_image)

@plugin_function(categories=["combine", "transform", "in assistant", "bia-bob-suggestion"])
def concatenate_along_y(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Concatenate two images or stacks along the Y axis.

    Parameters
    ----------
    input_image0: Image 
        First input image.
    input_image1: Image 
        Second input image.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_combineVertically
    """
    return clic._concatenate_along_y(device, input_image0, input_image1, output_image)

@plugin_function(categories=["combine", "transform", "in assistant", "bia-bob-suggestion"])
def concatenate_along_z(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Concatenate two images or stacks along the Z axis.

    Parameters
    ----------
    input_image0: Image 
        First input image.
    input_image1: Image 
        Second input image.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_concatenateStacks
    """
    return clic._concatenate_along_z(device, input_image0, input_image1, output_image)

@plugin_function
def count_touching_neighbors(
    input_image: Image,
    output_image: Optional[Image] =None,
    ignore_background: bool =True,
    device: Optional[Device] =None
) -> Image:
    """Takes a touch matrix as input and delivers a vector with number of touching
    neighbors per label as a vector. Note: Background is considered as something
    that can touch. To ignore touches with background, hand over a touch matrix
    where the first column (index = 0) has been set to 0. Use set_column for that.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    ignore_background: bool (= True)
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_countTouchingNeighbors
    """
    return clic._count_touching_neighbors(device, input_image, output_image, ignore_background)

@plugin_function
def crop_border(
    input_image: Image,
    output_image: Optional[Image] =None,
    border_size: int =1,
    device: Optional[Device] =None
) -> Image:
    """Crops an image by removing the outer pixels, per default 1. Notes * To make sure
    the output image has the right size, provide destination_image=None.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    border_size: int (= 1)
        Border size to crop.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._crop_border(device, input_image, output_image, int(border_size))

@plugin_function(categories=["filter", "background removal", "in assistant"])
def divide_by_gaussian_background(
    input_image: Image,
    output_image: Optional[Image] =None,
    sigma_x: float =2,
    sigma_y: float =2,
    sigma_z: float =2,
    device: Optional[Device] =None
) -> Image:
    """Applies Gaussian blur to the input image and divides the original by the result.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    sigma_x: float (= 2)
        Gaussian sigma value along x.
    sigma_y: float (= 2)
        Gaussian sigma value along y.
    sigma_z: float (= 2)
        Gaussian sigma value along z.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_divideByGaussianBackground
    """
    return clic._divide_by_gaussian_background(device, input_image, output_image, float(sigma_x), float(sigma_y), float(sigma_z))

@plugin_function
def degrees_to_radians(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Converts radians to degrees.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._degrees_to_radians(device, input_image, output_image)

@plugin_function(categories=["binarize", "in assistant"])
def detect_maxima_box(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =0,
    radius_y: int =0,
    radius_z: int =0,
    device: Optional[Device] =None
) -> Image:
    """Detects local maxima in a given square/cubic neighborhood. Pixels in the
    resulting image are set to 1 if there is no other pixel in a given radius which
    has a higher intensity, and to 0 otherwise.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 0)
        Radius along the x axis.
    radius_y: int (= 0)
        Radius along the y axis.
    radius_z: int (= 0)
        Radius along the z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_detectMaximaBox
    """
    return clic._detect_maxima_box(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z))

@plugin_function(categories=["binarize", "in assistant"])
def detect_maxima(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =0,
    radius_y: int =0,
    radius_z: int =0,
    connectivity: str ="box",
    device: Optional[Device] =None
) -> Image:
    """Detects local maxima in a given square/cubic neighborhood. Pixels in the
    resulting image are set to 1 if there is no other pixel in a given radius which
    has a higher intensity, and to 0 otherwise.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 0)
        Radius along the x axis.
    radius_y: int (= 0)
        Radius along the y axis.
    radius_z: int (= 0)
        Radius along the z axis.
    connectivity: str (= "box")
        Element shape, "box" or "sphere"
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_detectMaximaBox
    [2] https://clij.github.io/clij2-docs/reference_detectMaximaSphere
    """
    return clic._detect_maxima(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z), str(connectivity))

@plugin_function(categories=["binarize", "in assistant"])
def detect_minima_box(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =0,
    radius_y: int =0,
    radius_z: int =0,
    device: Optional[Device] =None
) -> Image:
    """Detects local maxima in a given square/cubic neighborhood. Pixels in the
    resulting image are set to 1 if there is no other pixel in a given radius which
    has a lower intensity, and to 0 otherwise.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 0)
        Radius along the x axis.
    radius_y: int (= 0)
        Radius along the y axis.
    radius_z: int (= 0)
        Radius along the z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_detectMinimaBox
    """
    return clic._detect_minima_box(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z))

@plugin_function(categories=["binarize", "in assistant"])
def detect_minima(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =0,
    radius_y: int =0,
    radius_z: int =0,
    connectivity: str ="box",
    device: Optional[Device] =None
) -> Image:
    """Detects local maxima in a given square/cubic neighborhood. Pixels in the
    resulting image are set to 1 if there is no other pixel in a given radius which
    has a lower intensity, and to 0 otherwise.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 0)
        Radius along the x axis.
    radius_y: int (= 0)
        Radius along the y axis.
    radius_z: int (= 0)
        Radius along the z axis.
    connectivity: str (= "box")
        Element shape, "box" or "sphere"
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_detectMinimaBox
    [2] https://clij.github.io/clij2-docs/reference_detectMinimaSphere
    """
    return clic._detect_minima(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z), str(connectivity))

@plugin_function(categories=["filter", "background removal", "bia-bob-suggestion"])
def difference_of_gaussian(
    input_image: Image,
    output_image: Optional[Image] =None,
    sigma1_x: float =2,
    sigma1_y: float =2,
    sigma1_z: float =2,
    sigma2_x: float =2,
    sigma2_y: float =2,
    sigma2_z: float =2,
    device: Optional[Device] =None
) -> Image:
    """Applies Gaussian blur to the input image twice with different sigma values
    resulting in two images which are then subtracted from each other. It is
    recommended to apply this operation to images of type Float (32 bit) as results
    might be negative.

    Parameters
    ----------
    input_image: Image 
        The input image to be processed.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    sigma1_x: float (= 2)
        Sigma of the first Gaussian filter in x
    sigma1_y: float (= 2)
        Sigma of the first Gaussian filter in y
    sigma1_z: float (= 2)
        Sigma of the first Gaussian filter in z
    sigma2_x: float (= 2)
        Sigma of the second Gaussian filter in x
    sigma2_y: float (= 2)
        Sigma of the second Gaussian filter in y
    sigma2_z: float (= 2)
        Sigma of the second Gaussian filter in z
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_differenceOfGaussian3D
    """
    return clic._difference_of_gaussian(device, input_image, output_image, float(sigma1_x), float(sigma1_y), float(sigma1_z), float(sigma2_x), float(sigma2_y), float(sigma2_z))

@plugin_function(categories=["label processing", "in assistant"])
def extend_labeling_via_voronoi(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Takes a label map image and dilates the regions using a octagon shape until they
    touch. The resulting label map is written to the output.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_extendLabelingViaVoronoi
    """
    return clic._extend_labeling_via_voronoi(device, input_image, output_image)

@plugin_function(categories=["filter"])
def invert(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Computes the negative value of all pixels in a given image. It is recommended to
    convert images to 32bit float before applying this operation. <pre>f(x) =
    x</pre> For binary images, use binaryNot.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_invert
    """
    return clic._invert(device, input_image, output_image)

@plugin_function(categories=["label", "in assistant"])
def label_spots(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Transforms a binary image with single pixles set to 1 to a labelled spots image.
    Transforms a spots image as resulting from maximum/minimum detection in an image
    of the same size where every spot has a number 1, 2,... n.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_labelSpots
    """
    return clic._label_spots(device, input_image, output_image)

@plugin_function(categories=["filter", "in assistant"])
def large_hessian_eigenvalue(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines the Hessian eigenvalues and returns the large eigenvalue image.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._large_hessian_eigenvalue(device, input_image, output_image)

@plugin_function
def maximum_of_all_pixels(
    input_image: Image,
    device: Optional[Device] =None
) -> float:
    """Determines the maximum of all pixels in a given image. It will be stored in a
    new row of ImageJs Results table in the column 'Max'.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    float

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_maximumOfAllPixels
    """
    return clic._maximum_of_all_pixels(device, input_image)

@plugin_function
def minimum_of_all_pixels(
    input_image: Image,
    device: Optional[Device] =None
) -> float:
    """Determines the minimum of all pixels in a given image. It will be stored in a
    new row of ImageJs Results table in the column 'Min'.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    float

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_minimumOfAllPixels
    """
    return clic._minimum_of_all_pixels(device, input_image)

@plugin_function
def minimum_of_masked_pixels(
    input_image: Image,
    mask: Image,
    device: Optional[Device] =None
) -> float:
    """Determines the minimum intensity in a masked image. But only in pixels which
    have nonzero values in another mask image.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    mask: Image 
        Input
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    float

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_minimumOfMaskedPixels
    """
    return clic._minimum_of_masked_pixels(device, input_image, mask)

@plugin_function(categories=["filter", "in assistant"])
def opening_box(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =0,
    radius_y: int =0,
    radius_z: int =0,
    device: Optional[Device] =None
) -> Image:
    """Opening operator, boxshaped Applies morphological opening to intensity images
    using a boxshaped footprint. This operator also works with binary images.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 0)
        Radius along the x axis.
    radius_y: int (= 0)
        Radius along the y axis.
    radius_z: int (= 0)
        Radius along the z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._opening_box(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z))

@plugin_function(categories=["filter", "in assistant"])
def opening_sphere(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: float =1,
    radius_y: float =1,
    radius_z: float =0,
    device: Optional[Device] =None
) -> Image:
    """Opening operator, sphereshaped Applies morphological opening to intensity images
    using a sphereshaped footprint. This operator also works with binary images.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: float (= 1)
        Radius along the x axis.
    radius_y: float (= 1)
        Radius along the y axis.
    radius_z: float (= 0)
        Radius along the z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._opening_sphere(device, input_image, output_image, float(radius_x), float(radius_y), float(radius_z))

@plugin_function(categories=["filter", "in assistant"])
def opening(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: float =1,
    radius_y: float =1,
    radius_z: float =0,
    connectivity: str ="box",
    device: Optional[Device] =None
) -> Image:
    """Opening operator, sphereshaped Applies morphological opening to intensity images
    using a sphereshaped footprint. This operator also works with binary images.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: float (= 1)
        Radius along the x axis.
    radius_y: float (= 1)
        Radius along the y axis.
    radius_z: float (= 0)
        Radius along the z axis.
    connectivity: str (= "box")
        Element shape, "box" or "sphere"
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._opening(device, input_image, output_image, float(radius_x), float(radius_y), float(radius_z), str(connectivity))

@plugin_function
def radians_to_degrees(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Converts radians to degrees

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._radians_to_degrees(device, input_image, output_image)

@plugin_function(categories=["label processing", "in assistant", "bia-bob-suggestion"])
def reduce_labels_to_label_edges(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Takes a label map and reduces all labels to their edges. Label IDs stay and
    background will be zero.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_reduceLabelsToLabelEdges
    """
    return clic._reduce_labels_to_label_edges(device, input_image, output_image)

@plugin_function(categories=["filter", "in assistant"])
def small_hessian_eigenvalue(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines the Hessian eigenvalues and returns the small eigenvalue image.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._small_hessian_eigenvalue(device, input_image, output_image)

@plugin_function(categories=["filter"])
def square(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Return the elementwise square of the input. This function is supposed to be
    similar to its counterpart in numpy [1]

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://numpy.org/doc/stable/reference/generated/numpy.square.html
    """
    return clic._square(device, input_image, output_image)

@plugin_function(categories=["combine", "in assistant"])
def squared_difference(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Determines the squared difference pixel by pixel between two images.

    Parameters
    ----------
    input_image0: Image 
        First input image.
    input_image1: Image 
        Second input image.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_squaredDifference
    """
    return clic._squared_difference(device, input_image0, input_image1, output_image)

@plugin_function(categories=["filter", "edge detection", "in assistant"])
def standard_deviation_box(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =1,
    radius_y: int =1,
    radius_z: int =1,
    device: Optional[Device] =None
) -> Image:
    """Computes the local standard deviation of a pixels box neighborhood. The box size
    is specified by its halfwidth, halfheight and halfdepth (radius). If 2D images
    are given, radius_z will be ignored.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 1)
        Radius along the x axis.
    radius_y: int (= 1)
        Radius along the y axis.
    radius_z: int (= 1)
        Radius along the z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_standardDeviationBox
    """
    return clic._standard_deviation_box(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z))

@plugin_function(categories=["filter", "edge detection", "in assistant"])
def standard_deviation_sphere(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =1,
    radius_y: int =1,
    radius_z: int =1,
    device: Optional[Device] =None
) -> Image:
    """Computes the local standard deviation of a pixels sphere neighborhood. The box
    size is specified by its halfwidth, halfheight and halfdepth (radius). If 2D
    images are given, radius_z will be ignored.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 1)
        Radius along the x axis.
    radius_y: int (= 1)
        Radius along the y axis.
    radius_z: int (= 1)
        Radius along the z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_standardDeviationSphere
    """
    return clic._standard_deviation_sphere(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z))

@plugin_function(categories=["filter", "edge detection", "in assistant"])
def standard_deviation(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =1,
    radius_y: int =1,
    radius_z: int =1,
    connectivity: str ="box",
    device: Optional[Device] =None
) -> Image:
    """Computes the local standard deviation of a pixels sphere neighborhood. The box
    size is specified by its halfwidth, halfheight and halfdepth (radius). If 2D
    images are given, radius_z will be ignored.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    radius_x: int (= 1)
        Radius along the x axis.
    radius_y: int (= 1)
        Radius along the y axis.
    radius_z: int (= 1)
        Radius along the z axis.
    connectivity: str (= "box")
        Neigborhood shape, "box" or "sphere"
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_standardDeviationBox
    [2] https://clij.github.io/clij2-docs/reference_standardDeviationSphere
    """
    return clic._standard_deviation(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z), str(connectivity))

@plugin_function(categories=["filter", "background removal", "in assistant", "bia-bob-suggestion"])
def subtract_gaussian_background(
    input_image: Image,
    output_image: Optional[Image] =None,
    sigma_x: float =2,
    sigma_y: float =2,
    sigma_z: float =2,
    device: Optional[Device] =None
) -> Image:
    """Applies Gaussian blur to the input image and subtracts the result from the
    original.

    Parameters
    ----------
    input_image: Image 
        Input image to process.
    output_image: Optional[Image] (= None)
        Output result image.
    sigma_x: float (= 2)
        Radius along the x axis.
    sigma_y: float (= 2)
        Radius along the y axis.
    sigma_z: float (= 2)
        Radius along the z axis.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_subtractGaussianBackground
    """
    return clic._subtract_gaussian_background(device, input_image, output_image, float(sigma_x), float(sigma_y), float(sigma_z))

@plugin_function(categories=["combine", "in assistant"])
def subtract_images(
    input_image0: Image,
    input_image1: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Subtracts one image X from another image Y pixel wise. <pre>f(x, y) = x y</pre>

    Parameters
    ----------
    input_image0: Image 
        First input image.
    input_image1: Image 
        Second input image.
    output_image: Optional[Image] (= None)
        Output result image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_subtractImages
    """
    return clic._subtract_images(device, input_image0, input_image1, output_image)

@plugin_function(categories=["transform", "in assistant"])
def sub_stack(
    input_image: Image,
    output_image: Optional[Image] =None,
    start_z: int =0,
    end_z: int =0,
    device: Optional[Device] =None
) -> Image:
    """Crop a volume into a new volume, along the z-axis.

    Parameters
    ----------
    input_image: Image 
        Input image.
    output_image: Optional[Image] (= None)
        Output image.
    start_z: int (= 0)
        Start z coordinate of the crop.
    end_z: int (= 0)
        End z coordinate of the crop.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_subStack
    """
    return clic._sub_stack(device, input_image, output_image, int(start_z), int(end_z))

@plugin_function(categories=["transform", "in assistant"])
def reduce_stack(
    input_image: Image,
    output_image: Optional[Image] =None,
    reduction_factor: int =2,
    offset: int =0,
    device: Optional[Device] =None
) -> Image:
    """Reduces the number of z-slices in a stack by a given factor. With the offset you
    have control which slices stays: with a factor 3 and offset 0, slices 0,3,6,
    etc. are kept. with a factor 4 and offset 1, slices 1,5,9, etc. are kept.

    Parameters
    ----------
    input_image: Image 
        Input image.
    output_image: Optional[Image] (= None)
        Output image.
    reduction_factor: int (= 2)
        Reduction factor.
    offset: int (= 0)
        Offset.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_reduceStack
    """
    return clic._reduce_stack(device, input_image, output_image, int(reduction_factor), int(offset))

@plugin_function
def sum_of_all_pixels(
    input_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> float:
    """Determines the sum of all pixels in a given image. It will be stored in a new
    row of ImageJs Results table in the column 'Sum'.

    Parameters
    ----------
    input_image: Optional[Image] (= None)
        Input image to process.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    float

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_sumOfAllPixels
    """
    return clic._sum_of_all_pixels(device, input_image)

@plugin_function(categories=["filter", "background removal", "in assistant"])
def top_hat_box(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: int =1,
    radius_y: int =1,
    radius_z: int =1,
    device: Optional[Device] =None
) -> Image:
    """Applies a tophat filter for background subtraction to the input image.

    Parameters
    ----------
    input_image: Image 
        The input image where the background is subtracted from.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    radius_x: int (= 1)
        Radius of the background determination region in X.
    radius_y: int (= 1)
        Radius of the background determination region in Y.
    radius_z: int (= 1)
        Radius of the background determination region in Z.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_topHatBox
    """
    return clic._top_hat_box(device, input_image, output_image, int(radius_x), int(radius_y), int(radius_z))

@plugin_function(categories=["filter", "background removal", "in assistant", "bia-bob-suggestion"])
def top_hat_sphere(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: float =1,
    radius_y: float =1,
    radius_z: float =1,
    device: Optional[Device] =None
) -> Image:
    """Applies a tophat filter for background subtraction to the input image.

    Parameters
    ----------
    input_image: Image 
        The input image where the background is subtracted from.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    radius_x: float (= 1)
        Radius of the background determination region in X.
    radius_y: float (= 1)
        Radius of the background determination region in Y.
    radius_z: float (= 1)
        Radius of the background determination region in Z.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_topHatSphere
    """
    return clic._top_hat_sphere(device, input_image, output_image, float(radius_x), float(radius_y), float(radius_z))

@plugin_function(categories=["filter", "background removal", "in assistant"])
def top_hat(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius_x: float =1,
    radius_y: float =1,
    radius_z: float =1,
    connectivity: str ="box",
    device: Optional[Device] =None
) -> Image:
    """Applies a tophat filter for background subtraction to the input image.

    Parameters
    ----------
    input_image: Image 
        The input image where the background is subtracted from.
    output_image: Optional[Image] (= None)
        The output image where results are written into.
    radius_x: float (= 1)
        Radius of the background determination region in X.
    radius_y: float (= 1)
        Radius of the background determination region in Y.
    radius_z: float (= 1)
        Radius of the background determination region in Z.
    connectivity: str (= "box")
        Element shape, "box" or "sphere"
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_topHatBox
    [2] https://clij.github.io/clij2-docs/reference_topHatSphere
    """
    return clic._top_hat(device, input_image, output_image, float(radius_x), float(radius_y), float(radius_z), str(connectivity))


import numpy as np

import pyclesperanto as cle


def test_zeros_arrays():
    gpu_arr = cle.Array.zeros((10, 10))

    assert gpu_arr.shape == (10, 10)
    assert gpu_arr.sum() == 0
    assert gpu_arr.max() == 0
    assert gpu_arr.min() == 0


def test_empty_like():
    arr = np.zeros((10, 10))
    gpu_arr = cle.Array.empty_like(arr)

    assert gpu_arr.shape == arr.shape


def test_zeros_like():
    arr = np.zeros((10, 10))
    gpu_arr = cle.Array.zeros_like(arr)

    assert gpu_arr.shape == arr.shape
    assert gpu_arr.sum() == 0
    assert gpu_arr.max() == 0
    assert gpu_arr.min() == 0


def test_to_device():
    arr = np.zeros((10, 10))
    gpu_arr = cle.Array.to_device(arr)

    assert gpu_arr.shape == arr.shape
    assert gpu_arr.sum() == 0
    assert gpu_arr.max() == 0
    assert gpu_arr.min() == 0


def test_from_array():
    arr = np.zeros((10, 10))
    gpu_arr = cle.Array.from_array(arr)

    assert gpu_arr.shape == arr.shape
    assert gpu_arr.sum() == 0
    assert gpu_arr.max() == 0
    assert gpu_arr.min() == 0


def test_array_ufunc_add():
    array1 = cle.push(np.ones((10, 10)))
    array2 = cle.push(np.ones((10, 10)))
    result = np.add(array1, array2)  # This should call array1.__array_ufunc__

    res = cle.pull(result)
    assert np.all(res == 2)


def test_repr():
    array = cle.Array.zeros((10, 10))
    result = repr(array)

    assert len(result) > 0
